@@ -24,11 +24,16 @@ ALLEGRO_DEBUG_CHANNEL("audio")
 #define MAX_LAG   (3)
 
 
-static void maybe_lock_mutex(ALLEGRO_MUTEX *mutex)
+/*
+ * To avoid deadlocks, unlock the mutex returned by this function, rather than
+ * whatever you passed as the argument.
+ */
+static ALLEGRO_MUTEX *maybe_lock_mutex(ALLEGRO_MUTEX *mutex)
 {
    if (mutex) {
       al_lock_mutex(mutex);
    }
+   return mutex;
 }
 
 
@@ -121,10 +126,10 @@ ALLEGRO_AUDIO_STREAM *al_create_audio_stream(size_t fragment_count,
    }
 
    for (i = 0; i < fragment_count; i++) {
-      stream->pending_bufs[i] =
-         (char *) stream->main_buffer
-         + i * (MAX_LAG*bytes_per_sample + bytes_per_frag_buf)
-         + MAX_LAG*bytes_per_sample;
+      char *buffer = (char *)stream->main_buffer
+         + i * (MAX_LAG * bytes_per_sample + bytes_per_frag_buf);
+      al_fill_silence(buffer, MAX_LAG, depth, chan_conf);
+      stream->used_bufs[i] = buffer + MAX_LAG * bytes_per_sample;
    }
 
    al_init_user_event_source(&stream->spl.es);
@@ -302,6 +307,26 @@ bool al_get_audio_stream_attached(const ALLEGRO_AUDIO_STREAM *stream)
    return (stream->spl.parent.u.ptr != NULL);
 }
 
+/* Function: al_get_audio_stream_played_samples
+*/
+uint64_t al_get_audio_stream_played_samples(const ALLEGRO_AUDIO_STREAM *stream)
+{
+   uint64_t result;
+   ALLEGRO_MUTEX *stream_mutex;
+   ASSERT(stream);
+
+   stream_mutex = maybe_lock_mutex(stream->spl.mutex);
+   if (stream->spl.spl_data.buffer.ptr) {
+      result = stream->consumed_fragments * stream->spl.spl_data.len +
+         stream->spl.pos;
+   }
+   else {
+      result = 0;
+   }
+   maybe_unlock_mutex(stream_mutex);
+
+   return result;
+}
 
 /* Function: al_get_audio_stream_fragment
 */
@@ -309,9 +334,10 @@ void *al_get_audio_stream_fragment(const ALLEGRO_AUDIO_STREAM *stream)
 {
    size_t i;
    void *fragment;
+   ALLEGRO_MUTEX *stream_mutex;
    ASSERT(stream);
 
-   maybe_lock_mutex(stream->spl.mutex);
+   stream_mutex = maybe_lock_mutex(stream->spl.mutex);
 
    if (!stream->used_bufs[0]) {
       /* No free fragments are available. */
@@ -325,7 +351,7 @@ void *al_get_audio_stream_fragment(const ALLEGRO_AUDIO_STREAM *stream)
       stream->used_bufs[i] = NULL;
    }
 
-   maybe_unlock_mutex(stream->spl.mutex);
+   maybe_unlock_mutex(stream_mutex);
 
    return fragment;
 }
@@ -352,8 +378,7 @@ bool al_set_audio_stream_speed(ALLEGRO_AUDIO_STREAM *stream, float val)
    stream->spl.speed = val;
    if (stream->spl.parent.u.mixer) {
       ALLEGRO_MIXER *mixer = stream->spl.parent.u.mixer;
-
-      maybe_lock_mutex(stream->spl.mutex);
+      ALLEGRO_MUTEX *stream_mutex = maybe_lock_mutex(stream->spl.mutex);
 
       stream->spl.step = (stream->spl.spl_data.frequency) * stream->spl.speed;
       stream->spl.step_denom = mixer->ss.spl_data.frequency;
@@ -362,7 +387,7 @@ bool al_set_audio_stream_speed(ALLEGRO_AUDIO_STREAM *stream, float val)
          stream->spl.step = 1;
       }
 
-      maybe_unlock_mutex(stream->spl.mutex);
+      maybe_unlock_mutex(stream_mutex);
    }
 
    return true;
@@ -389,10 +414,9 @@ bool al_set_audio_stream_gain(ALLEGRO_AUDIO_STREAM *stream, float val)
        */
       if (stream->spl.parent.u.mixer) {
          ALLEGRO_MIXER *mixer = stream->spl.parent.u.mixer;
-
-         maybe_lock_mutex(stream->spl.mutex);
+         ALLEGRO_MUTEX *stream_mutex = maybe_lock_mutex(stream->spl.mutex);
          _al_kcm_mixer_rejig_sample_matrix(mixer, &stream->spl);
-         maybe_unlock_mutex(stream->spl.mutex);
+         maybe_unlock_mutex(stream_mutex);
       }
    }
 
@@ -424,10 +448,9 @@ bool al_set_audio_stream_pan(ALLEGRO_AUDIO_STREAM *stream, float val)
        */
       if (stream->spl.parent.u.mixer) {
          ALLEGRO_MIXER *mixer = stream->spl.parent.u.mixer;
-
-         maybe_lock_mutex(stream->spl.mutex);
+         ALLEGRO_MUTEX *stream_mutex = maybe_lock_mutex(stream->spl.mutex);
          _al_kcm_mixer_rejig_sample_matrix(mixer, &stream->spl);
-         maybe_unlock_mutex(stream->spl.mutex);
+         maybe_unlock_mutex(stream_mutex);
       }
    }
 
@@ -461,35 +484,84 @@ bool al_set_audio_stream_playmode(ALLEGRO_AUDIO_STREAM *stream,
 }
 
 
+static void reset_stopped_stream(ALLEGRO_AUDIO_STREAM *stream)
+{
+   const int bytes_per_sample =
+      al_get_channel_count(stream->spl.spl_data.chan_conf) *
+      al_get_audio_depth_size(stream->spl.spl_data.depth);
+   const int fragment_buffer_size =
+      bytes_per_sample * (stream->spl.spl_data.len + MAX_LAG);
+   size_t i, n;
+
+   /* Write silence to the "invisible" part in between fragment buffers to
+    * avoid interpolation artifacts.  It's tempting to zero the complete
+    * memory block in one go but some of the buffers might be getting
+    * refilled.  So they are currently "owned" by the library user and
+    * should not be overwritten.  But zeroing the parts not visible to the
+    * user should be OK.
+    */
+   for (i = 0; i < stream->buf_count; ++i) {
+      al_fill_silence((char *)stream->main_buffer + i * fragment_buffer_size,
+         MAX_LAG, stream->spl.spl_data.depth, stream->spl.spl_data.chan_conf);
+   }
+
+   /* Get the current number of entries in the used_buf list. */
+   for (n = 0; n < stream->buf_count && stream->used_bufs[n]; n++)
+      ;
+
+   /* Move everything from pending_bufs to used_bufs. */
+   i = 0;
+   while (i < stream->buf_count &&
+         n < stream->buf_count &&
+         stream->pending_bufs[i])
+   {
+      stream->used_bufs[n] = stream->pending_bufs[i];
+      stream->pending_bufs[i] = NULL;
+      n++;
+      i++;
+   }
+
+   /* No fragment buffer is currently playing. */
+   stream->spl.spl_data.buffer.ptr = NULL;
+   stream->spl.pos = stream->spl.spl_data.len;
+   stream->spl.pos_bresenham_error = 0;
+   stream->consumed_fragments = 0;
+}
+
+
 /* Function: al_set_audio_stream_playing
  */
 bool al_set_audio_stream_playing(ALLEGRO_AUDIO_STREAM *stream, bool val)
 {
+   bool rc = true;
+   ALLEGRO_MUTEX *stream_mutex;
    ASSERT(stream);
 
    if (stream->spl.parent.u.ptr && stream->spl.parent.is_voice) {
       ALLEGRO_VOICE *voice = stream->spl.parent.u.voice;
-      bool rc;
-
-      if (val == stream->spl.is_playing) {
-         return true;
+      if (val != stream->spl.is_playing) {
+         rc = _al_kcm_set_voice_playing(voice, voice->mutex, val);
       }
-
-      rc = _al_kcm_set_voice_playing(voice, val);
-      if (rc) {
-         stream->spl.is_playing = val;
-      }
-      return rc;
    }
 
-   stream->spl.is_playing = val;
+   stream_mutex = maybe_lock_mutex(stream->spl.mutex);
 
-   if (!val) {
-      maybe_lock_mutex(stream->spl.mutex);
-      stream->spl.pos = stream->spl.spl_data.len;
-      maybe_unlock_mutex(stream->spl.mutex);
+   stream->spl.is_playing = rc && val;
+
+   if (stream->spl.is_playing) {
+      /* We usually emit fragment events when a fragment is used up, but if the
+       * stream has zero pending fragments to begin with then no events would
+       * ever be emitted.
+       */
+      _al_kcm_emit_stream_events(stream);
    }
-   return true;
+   else if (!val) {
+      reset_stopped_stream(stream);
+   }
+
+   maybe_unlock_mutex(stream_mutex);
+
+   return rc;
 }
 
 
@@ -511,9 +583,10 @@ bool al_set_audio_stream_fragment(ALLEGRO_AUDIO_STREAM *stream, void *val)
 {
    size_t i;
    bool ret;
+   ALLEGRO_MUTEX *stream_mutex;
    ASSERT(stream);
 
-   maybe_lock_mutex(stream->spl.mutex);
+   stream_mutex = maybe_lock_mutex(stream->spl.mutex);
 
    for (i = 0; i < stream->buf_count && stream->pending_bufs[i] ; i++)
       ;
@@ -527,7 +600,7 @@ bool al_set_audio_stream_fragment(ALLEGRO_AUDIO_STREAM *stream, void *val)
       ret = false;
    }
 
-   maybe_unlock_mutex(stream->spl.mutex);
+   maybe_unlock_mutex(stream_mutex);
 
    return ret;
 }
@@ -580,6 +653,8 @@ bool _al_kcm_refill_stream(ALLEGRO_AUDIO_STREAM *stream)
          (char *) new_buf - bytes_per_sample * MAX_LAG,
          (char *) old_buf + bytes_per_sample * (spl->pos-MAX_LAG),
          bytes_per_sample * MAX_LAG);
+
+      stream->consumed_fragments++;
    }
 
    stream->spl.pos = 0;
@@ -604,6 +679,11 @@ void *_al_kcm_feed_stream(ALLEGRO_THREAD *self, void *vstream)
    queue = al_create_event_queue();
    al_register_event_source(queue, &stream->spl.es);
 
+   al_lock_mutex(stream->feed_thread_started_mutex);
+   stream->feed_thread_started = true;
+   al_broadcast_cond(stream->feed_thread_started_cond);
+   al_unlock_mutex(stream->feed_thread_started_mutex);
+
    stream->quit_feed_thread = false;
 
    while (!stream->quit_feed_thread) {
@@ -616,6 +696,7 @@ void *_al_kcm_feed_stream(ALLEGRO_THREAD *self, void *vstream)
           && !stream->is_draining) {
          unsigned long bytes;
          unsigned long bytes_written;
+         ALLEGRO_MUTEX *stream_mutex;
 
          fragment = al_get_audio_stream_fragment(stream);
          if (!fragment) {
@@ -627,24 +708,30 @@ void *_al_kcm_feed_stream(ALLEGRO_THREAD *self, void *vstream)
                al_get_channel_count(stream->spl.spl_data.chan_conf) *
                al_get_audio_depth_size(stream->spl.spl_data.depth);
 
-         maybe_lock_mutex(stream->spl.mutex);
+         stream_mutex = maybe_lock_mutex(stream->spl.mutex);
          bytes_written = stream->feeder(stream, fragment, bytes);
-         maybe_unlock_mutex(stream->spl.mutex);
+         maybe_unlock_mutex(stream_mutex);
 
-        /* In case it reaches the end of the stream source, stream feeder will
-         * fill the remaining space with silence. If we should loop, rewind the
-         * stream and override the silence with the beginning.
-         * In extreme cases we need to repeat it multiple times.
-         */
-         while (bytes_written < bytes &&
-                  stream->spl.loop == _ALLEGRO_PLAYMODE_STREAM_ONEDIR) {
-            size_t bw;
-            al_rewind_audio_stream(stream);
-            maybe_lock_mutex(stream->spl.mutex);
-            bw = stream->feeder(stream, fragment + bytes_written,
-               bytes - bytes_written);
-            bytes_written += bw;
-            maybe_unlock_mutex(stream->spl.mutex);
+         if (stream->spl.loop == _ALLEGRO_PLAYMODE_STREAM_ONEDIR) {
+            /* Keep rewinding until the fragment is filled. */
+            while (bytes_written < bytes &&
+                     stream->spl.loop == _ALLEGRO_PLAYMODE_STREAM_ONEDIR) {
+               size_t bw;
+               al_rewind_audio_stream(stream);
+               stream_mutex = maybe_lock_mutex(stream->spl.mutex);
+               bw = stream->feeder(stream, fragment + bytes_written,
+                  bytes - bytes_written);
+               bytes_written += bw;
+               maybe_unlock_mutex(stream_mutex);
+            }
+         }
+         else if (bytes_written < bytes) {
+            /* Fill the rest of the fragment with silence. */
+            int silence_samples = (bytes - bytes_written) /
+               (al_get_channel_count(stream->spl.spl_data.chan_conf) *
+                al_get_audio_depth_size(stream->spl.spl_data.depth));
+            al_fill_silence(fragment + bytes_written, silence_samples,
+                            stream->spl.spl_data.depth, stream->spl.spl_data.chan_conf);
          }
 
          if (!al_set_audio_stream_fragment(stream, fragment)) {
@@ -707,9 +794,9 @@ bool al_rewind_audio_stream(ALLEGRO_AUDIO_STREAM *stream)
    bool ret;
 
    if (stream->rewind_feeder) {
-      maybe_lock_mutex(stream->spl.mutex);
+      ALLEGRO_MUTEX *stream_mutex = maybe_lock_mutex(stream->spl.mutex);
       ret = stream->rewind_feeder(stream);
-      maybe_unlock_mutex(stream->spl.mutex);
+      maybe_unlock_mutex(stream_mutex);
       return ret;
    }
 
@@ -724,9 +811,9 @@ bool al_seek_audio_stream_secs(ALLEGRO_AUDIO_STREAM *stream, double time)
    bool ret;
 
    if (stream->seek_feeder) {
-      maybe_lock_mutex(stream->spl.mutex);
+      ALLEGRO_MUTEX *stream_mutex = maybe_lock_mutex(stream->spl.mutex);
       ret = stream->seek_feeder(stream, time);
-      maybe_unlock_mutex(stream->spl.mutex);
+      maybe_unlock_mutex(stream_mutex);
       return ret;
    }
 
@@ -741,9 +828,9 @@ double al_get_audio_stream_position_secs(ALLEGRO_AUDIO_STREAM *stream)
    double ret;
 
    if (stream->get_feeder_position) {
-      maybe_lock_mutex(stream->spl.mutex);
+      ALLEGRO_MUTEX *stream_mutex = maybe_lock_mutex(stream->spl.mutex);
       ret = stream->get_feeder_position(stream);
-      maybe_unlock_mutex(stream->spl.mutex);
+      maybe_unlock_mutex(stream_mutex);
       return ret;
    }
 
@@ -758,9 +845,9 @@ double al_get_audio_stream_length_secs(ALLEGRO_AUDIO_STREAM *stream)
    double ret;
 
    if (stream->get_feeder_length) {
-      maybe_lock_mutex(stream->spl.mutex);
+      ALLEGRO_MUTEX *stream_mutex = maybe_lock_mutex(stream->spl.mutex);
       ret = stream->get_feeder_length(stream);
-      maybe_unlock_mutex(stream->spl.mutex);
+      maybe_unlock_mutex(stream_mutex);
       return ret;
    }
 
@@ -779,9 +866,9 @@ bool al_set_audio_stream_loop_secs(ALLEGRO_AUDIO_STREAM *stream,
       return false;
 
    if (stream->set_feeder_loop) {
-      maybe_lock_mutex(stream->spl.mutex);
+      ALLEGRO_MUTEX *stream_mutex = maybe_lock_mutex(stream->spl.mutex);
       ret = stream->set_feeder_loop(stream, start, end);
-      maybe_unlock_mutex(stream->spl.mutex);
+      maybe_unlock_mutex(stream_mutex);
       return ret;
    }
 
