@@ -24,16 +24,29 @@ bool al_init_image_addon(void)
    success |= al_register_bitmap_saver(".pcx", _al_save_pcx);
    success |= al_register_bitmap_loader_f(".pcx", _al_load_pcx_f);
    success |= al_register_bitmap_saver_f(".pcx", _al_save_pcx_f);
+   success |= al_register_bitmap_identifier(".pcx", _al_identify_pcx);
 
    success |= al_register_bitmap_loader(".bmp", _al_load_bmp);
    success |= al_register_bitmap_saver(".bmp", _al_save_bmp);
    success |= al_register_bitmap_loader_f(".bmp", _al_load_bmp_f);
    success |= al_register_bitmap_saver_f(".bmp", _al_save_bmp_f);
+   success |= al_register_bitmap_identifier(".bmp", _al_identify_bmp);
 
    success |= al_register_bitmap_loader(".tga", _al_load_tga);
    success |= al_register_bitmap_saver(".tga", _al_save_tga);
    success |= al_register_bitmap_loader_f(".tga", _al_load_tga_f);
    success |= al_register_bitmap_saver_f(".tga", _al_save_tga_f);
+   success |= al_register_bitmap_identifier(".tga", _al_identify_tga);
+
+   success |= al_register_bitmap_loader(".dds", _al_load_dds);
+   success |= al_register_bitmap_loader_f(".dds", _al_load_dds_f);
+   success |= al_register_bitmap_identifier(".dds", _al_identify_dds);
+
+   /* Even if we don't have libpng or libjpeg we most likely have a
+    * native reader for those instead so always identify them.
+    */
+   success |= al_register_bitmap_identifier(".png", _al_identify_png);
+   success |= al_register_bitmap_identifier(".jpg", _al_identify_jpg);
 
 /* ALLEGRO_CFG_IIO_HAVE_* is sufficient to know that the library
    should be used. i.e., ALLEGRO_CFG_IIO_HAVE_GDIPLUS and
@@ -52,12 +65,27 @@ bool al_init_image_addon(void)
             success |= al_register_bitmap_saver(extensions[i], _al_save_gdiplus_bitmap);
          }
 
-	     success |= al_register_bitmap_saver_f(".tif", _al_save_gdiplus_tif_f);
-	     success |= al_register_bitmap_saver_f(".tiff", _al_save_gdiplus_tif_f);
+         success |= al_register_bitmap_saver_f(".tif", _al_save_gdiplus_tif_f);
+         success |= al_register_bitmap_saver_f(".tiff", _al_save_gdiplus_tif_f);
          success |= al_register_bitmap_saver_f(".gif", _al_save_gdiplus_gif_f);
-	     success |= al_register_bitmap_saver_f(".png", _al_save_gdiplus_png_f);
-	     success |= al_register_bitmap_saver_f(".jpg", _al_save_gdiplus_jpg_f);
-	     success |= al_register_bitmap_saver_f(".jpeg", _al_save_gdiplus_jpg_f);	  
+         success |= al_register_bitmap_saver_f(".png", _al_save_gdiplus_png_f);
+         success |= al_register_bitmap_saver_f(".jpg", _al_save_gdiplus_jpg_f);
+         success |= al_register_bitmap_saver_f(".jpeg", _al_save_gdiplus_jpg_f);  
+      }
+   }
+#endif
+
+#ifdef ALLEGRO_CFG_IIO_HAVE_ANDROID
+   {
+      char const *extensions[] = {".webp", ".jpg", ".jpeg", ".ico", ".gif", 
+         ".wbmp", ".png", NULL};
+      int i;
+
+
+      for (i = 0; extensions[i]; i++) {
+         success |= al_register_bitmap_loader(extensions[i], _al_load_android_bitmap);
+         success |= al_register_bitmap_loader_f(extensions[i], _al_load_android_bitmap_f);
+         //success |= al_register_bitmap_saver(extensions[i], _al_save_android_bitmap);
       }
    }
 #endif
