@@ -106,6 +106,7 @@ static ALLEGRO_BITMAP *really_load_png(png_structp png_ptr, png_infop info_ptr,
    unsigned char *buf;
    unsigned char *dest;
    bool premul = !(flags & ALLEGRO_NO_PREMULTIPLIED_ALPHA);
+   bool index_only;
 
    ALLEGRO_ASSERT(png_ptr && info_ptr);
 
@@ -214,8 +215,18 @@ static ALLEGRO_BITMAP *really_load_png(png_structp png_ptr, png_infop info_ptr,
    else
       buf = al_malloc(real_rowbytes);
 
-   lock = al_lock_bitmap(bmp, ALLEGRO_PIXEL_FORMAT_ABGR_8888_LE,
-      ALLEGRO_LOCK_WRITEONLY);
+   if (bpp == 8 && (color_type & PNG_COLOR_MASK_PALETTE) &&
+      (flags & ALLEGRO_KEEP_INDEX))
+   {
+      lock = al_lock_bitmap(bmp, ALLEGRO_PIXEL_FORMAT_SINGLE_CHANNEL_8,
+         ALLEGRO_LOCK_WRITEONLY);
+      index_only = true;
+   }
+   else {
+      lock = al_lock_bitmap(bmp, ALLEGRO_PIXEL_FORMAT_ABGR_8888_LE,
+         ALLEGRO_LOCK_WRITEONLY);
+      index_only = false;
+   }
 
    /* Read the image, one line at a time (easier to debug!) */
    for (pass = 0; pass < number_passes; pass++) {
@@ -232,25 +243,33 @@ static ALLEGRO_BITMAP *really_load_png(png_structp png_ptr, png_infop info_ptr,
          if (interlace_type == PNG_INTERLACE_ADAM7)
             ptr = buf + y * real_rowbytes;
          else
-             ptr = buf;
+            ptr = buf;
          png_read_row(png_ptr, NULL, ptr);
    
          switch (bpp) {
             case 8:
-               if (color_type & PNG_COLOR_MASK_PALETTE) {
+               if (index_only) {
+                  for (i = 0; i < width; i++) {
+                     *(dest++) = *(ptr++);
+                  }
+               }
+               else if (color_type & PNG_COLOR_MASK_PALETTE) {
                   for (i = 0; i < width; i++) {
                      int pix = ptr[0];
-                     int ti;
                      ptr++;
                      dest[0] = pal[pix].r;
                      dest[1] = pal[pix].g;
                      dest[2] = pal[pix].b;
-                     dest[3] = 255;
-                     for (ti = 0; ti < num_trans; ti++) {
-                        if (trans[ti] == pix) {
-                           dest[0] = dest[1] = dest[2] = dest[3] = 0;
-                           break;
+                     if (pix < num_trans) {
+                        int a = trans[pix];
+                        dest[3] = a;
+                        if (premul) {
+                           dest[0] = dest[0] * a / 255;
+                           dest[1] = dest[1] * a / 255;
+                           dest[2] = dest[2] * a / 255;
                         }
+                     } else {
+                        dest[3] = 255;
                      }
                      dest += 4;
                   }
@@ -321,7 +340,7 @@ static ALLEGRO_BITMAP *really_load_png(png_structp png_ptr, png_infop info_ptr,
 
 /* Load a PNG file from disk, doing colour coversion if required.
  */
-static ALLEGRO_BITMAP *load_png_f(ALLEGRO_FILE *fp, int flags)
+ALLEGRO_BITMAP *_al_load_png_f(ALLEGRO_FILE *fp, int flags)
 {
    jmp_buf jmpbuf;
    ALLEGRO_BITMAP *bmp;
@@ -383,14 +402,9 @@ static ALLEGRO_BITMAP *load_png_f(ALLEGRO_FILE *fp, int flags)
 
 
 
-ALLEGRO_BITMAP *_al_load_png_f(ALLEGRO_FILE *fp)
-{
-   return load_png_f(fp, al_get_new_bitmap_flags());
-}
 
 
-
-ALLEGRO_BITMAP *_al_load_png(const char *filename)
+ALLEGRO_BITMAP *_al_load_png(const char *filename, int flags)
 {
    ALLEGRO_FILE *fp;
    ALLEGRO_BITMAP *bmp;
@@ -401,7 +415,7 @@ ALLEGRO_BITMAP *_al_load_png(const char *filename)
    if (!fp)
       return NULL;
 
-   bmp = _al_load_png_f(fp);
+   bmp = _al_load_png_f(fp, flags);
 
    al_fclose(fp);
 
@@ -549,7 +563,8 @@ bool _al_save_png_f(ALLEGRO_FILE *fp, ALLEGRO_BITMAP *bmp)
 bool _al_save_png(const char *filename, ALLEGRO_BITMAP *bmp)
 {
    ALLEGRO_FILE *fp;
-   bool result;
+   bool retsave;
+   bool retclose;
 
    ALLEGRO_ASSERT(filename);
    ALLEGRO_ASSERT(bmp);
@@ -560,10 +575,10 @@ bool _al_save_png(const char *filename, ALLEGRO_BITMAP *bmp)
       return false;
    }
 
-   result = _al_save_png_f(fp, bmp);
-   al_fclose(fp);
+   retsave = _al_save_png_f(fp, bmp);
+   retclose = al_fclose(fp);
 
-   return result;
+   return retsave && retclose;
 }
 
 /* vim: set sts=3 sw=3 et: */
