@@ -4,6 +4,7 @@
 #include "allegro5/internal/aintern_exitfunc.h"
 #include "allegro5/internal/aintern_native_dialog.h"
 #include "allegro5/internal/aintern_system.h"
+#include "allegro5/internal/aintern_vector.h"
 
 ALLEGRO_DEBUG_CHANNEL("native_dialog")
 
@@ -19,6 +20,7 @@ bool al_init_native_dialog_addon(void)
          return false;
       }
       inited_addon = true;
+      al_init_user_event_source(al_get_default_menu_event_source());
       _al_add_exit_func(al_shutdown_native_dialog_addon,
          "al_shutdown_native_dialog_addon");
    }
@@ -30,6 +32,7 @@ bool al_init_native_dialog_addon(void)
 void al_shutdown_native_dialog_addon(void)
 {
    if (inited_addon) {
+      al_destroy_user_event_source(al_get_default_menu_event_source());
       _al_shutdown_native_dialog_addon();
       inited_addon = false;
    }
@@ -53,7 +56,7 @@ ALLEGRO_FILECHOOSER *al_create_native_file_dialog(
    fc->fc_patterns = al_ustr_new(patterns);
    fc->flags = mode;
 
-   _al_register_destructor(_al_dtor_list, fc,
+   _al_register_destructor(_al_dtor_list, "native_dialog", fc,
       (void (*)(void *))al_destroy_native_file_dialog);
 
    return (ALLEGRO_FILECHOOSER *)fc;
@@ -118,6 +121,10 @@ int al_show_native_message_box(ALLEGRO_DISPLAY *display,
    ALLEGRO_NATIVE_DIALOG *fc;
    int r;
 
+   ASSERT(title);
+   ASSERT(heading);
+   ASSERT(text);
+
    /* Note: the message box code cannot assume that Allegro is installed.
     * al_malloc and ustr functions are okay (with the assumption that the
     * user doesn't change the memory management functions in another thread
@@ -143,7 +150,6 @@ int al_show_native_message_box(ALLEGRO_DISPLAY *display,
    return r;
 }
 
-
 /* Function: al_get_allegro_native_dialog_version
  */
 uint32_t al_get_allegro_native_dialog_version(void)
