@@ -16,8 +16,6 @@
  *      See readme.txt for copyright information.
  */
 
-#define ALLEGRO_INTERNAL_UNSTABLE
-
 #include "allegro5/allegro.h"
 #include "allegro5/allegro_primitives.h"
 #include "allegro5/internal/aintern_bitmap.h"
@@ -78,24 +76,24 @@ static bool is_legacy_card(void)
    return legacy_card;
 }
 
-struct LEGACY_VERTEX
+typedef struct LEGACY_VERTEX
 {
    float x, y, z;
    DWORD color;
    float u, v;
-};
+} LEGACY_VERTEX;
 
 static uint8_t* legacy_buffer;
 static size_t legacy_buffer_size = 0;
 #define A5V_FVF (D3DFVF_XYZ | D3DFVF_TEX2 | D3DFVF_TEXCOORDSIZE2(0) | D3DFVF_TEXCOORDSIZE4(1))
 #define A5V_LEGACY_FVF (D3DFVF_XYZ | D3DFVF_DIFFUSE | D3DFVF_TEX1)
 
-struct DISPLAY_LOCAL_DATA
+typedef struct DISPLAY_LOCAL_DATA
 {
    LPDIRECT3DDEVICE9 device;
    LPDIRECT3DVERTEXSHADER9 shader;
    ALLEGRO_INDEX_BUFFER* loop_index_buffer;
-};
+} DISPLAY_LOCAL_DATA;
 
 static DISPLAY_LOCAL_DATA* display_local_data;
 static int display_local_data_size = 0;
@@ -404,10 +402,32 @@ static D3D_STATE setup_state(LPDIRECT3DDEVICE9 device, const ALLEGRO_VERTEX_DECL
       }
    }
 
+   device->GetSamplerState(0, D3DSAMP_ADDRESSU, &state.old_wrap_state[0]);
+   device->GetSamplerState(0, D3DSAMP_ADDRESSV, &state.old_wrap_state[1]);
+
+   device->SetSamplerState(0, D3DSAMP_ADDRESSU, D3DTADDRESS_WRAP);
+   device->SetSamplerState(0, D3DSAMP_ADDRESSV, D3DTADDRESS_WRAP);
+
    if (texture) {
-      device->GetSamplerState(0, D3DSAMP_ADDRESSU, &state.old_wrap_state[0]);
-      device->GetSamplerState(0, D3DSAMP_ADDRESSV, &state.old_wrap_state[1]);
-      _al_set_d3d_sampler_state(device, 0, texture, true);
+      int texture_flags = al_get_bitmap_flags(texture);
+      if (texture_flags & ALLEGRO_MIN_LINEAR) {
+         device->SetSamplerState(0, D3DSAMP_MINFILTER, D3DTEXF_LINEAR);
+      }
+      else {
+         device->SetSamplerState(0, D3DSAMP_MINFILTER, D3DTEXF_POINT);
+      }
+      if (texture_flags & ALLEGRO_MAG_LINEAR) {
+         device->SetSamplerState(0, D3DSAMP_MAGFILTER, D3DTEXF_LINEAR);
+      }
+      else {
+         device->SetSamplerState(0, D3DSAMP_MAGFILTER, D3DTEXF_POINT);
+      }
+      if (texture_flags & ALLEGRO_MIPMAP) {
+         device->SetSamplerState(0, D3DSAMP_MIPFILTER, D3DTEXF_LINEAR);
+      }
+      else {
+         device->SetSamplerState(0, D3DSAMP_MIPFILTER, D3DTEXF_NONE);
+      }
    }
 
    return state;
@@ -426,10 +446,8 @@ static void revert_state(D3D_STATE state, LPDIRECT3DDEVICE9 device, ALLEGRO_BITM
    }
 #endif
 
-   if (texture) {
-      device->SetSamplerState(0, D3DSAMP_ADDRESSU, state.old_wrap_state[0]);
-      device->SetSamplerState(0, D3DSAMP_ADDRESSV, state.old_wrap_state[1]);
-   }
+   device->SetSamplerState(0, D3DSAMP_ADDRESSU, state.old_wrap_state[0]);
+   device->SetSamplerState(0, D3DSAMP_ADDRESSV, state.old_wrap_state[1]);
 
    if(!state.old_vtx_shader && is_legacy_card() && texture) {
       device->SetTextureStageState(0, D3DTSS_TEXTURETRANSFORMFLAGS, state.old_ttf_state);
@@ -476,7 +494,6 @@ static int draw_prim_raw(ALLEGRO_BITMAP* target, ALLEGRO_BITMAP* texture,
          return _al_draw_prim_indexed_soft(texture, vtx, decl, indices, num_vtx, type);
    }
 
-   int num_idx = num_vtx;
    if(indices)
    {
       int ii;
@@ -492,7 +509,6 @@ static int draw_prim_raw(ALLEGRO_BITMAP* target, ALLEGRO_BITMAP* texture,
             max_idx = idx;
          }
       }
-      num_idx = max_idx + 1 - min_idx;
    }
 
    device = al_get_d3d_device(disp);
@@ -573,40 +589,37 @@ static int draw_prim_raw(ALLEGRO_BITMAP* target, ALLEGRO_BITMAP* texture,
          switch (type) {
             case ALLEGRO_PRIM_LINE_LIST: {
                num_primitives = num_vtx / 2;
-               device->DrawIndexedPrimitiveUP(D3DPT_LINELIST, min_idx, num_idx, num_primitives, indices, D3DFMT_INDEX32, vtx, stride);
+               device->DrawIndexedPrimitiveUP(D3DPT_LINELIST, min_idx, max_idx + 1, num_primitives, indices, D3DFMT_INDEX32, vtx, stride);
                break;
             };
             case ALLEGRO_PRIM_LINE_STRIP: {
                num_primitives = num_vtx - 1;
-               device->DrawIndexedPrimitiveUP(D3DPT_LINESTRIP, min_idx, num_idx, num_primitives, indices, D3DFMT_INDEX32, vtx, stride);
+               device->DrawIndexedPrimitiveUP(D3DPT_LINESTRIP, min_idx, max_idx + 1, num_primitives, indices, D3DFMT_INDEX32, vtx, stride);
                break;
             };
             case ALLEGRO_PRIM_LINE_LOOP: {
                int in[2];
                num_primitives = num_vtx - 1;
-               device->DrawIndexedPrimitiveUP(D3DPT_LINESTRIP, min_idx, num_idx, num_primitives, indices, D3DFMT_INDEX32, vtx, stride);
+               device->DrawIndexedPrimitiveUP(D3DPT_LINESTRIP, min_idx, max_idx + 1, num_primitives, indices, D3DFMT_INDEX32, vtx, stride);
    
                in[0] = indices[0];
                in[1] = indices[num_vtx-1];
-               min_idx = in[0] > in[1] ? in[1] : in[0];
-               max_idx = in[0] > in[1] ? in[0] : in[1];
-               num_idx = max_idx - min_idx + 1;
-               device->DrawIndexedPrimitiveUP(D3DPT_LINELIST, min_idx, num_idx, 1, in, D3DFMT_INDEX32, vtx, stride);
+               device->DrawIndexedPrimitiveUP(D3DPT_LINELIST, min_idx, max_idx + 1, 1, in, D3DFMT_INDEX32, vtx, stride);
                break;
             };
             case ALLEGRO_PRIM_TRIANGLE_LIST: {
                num_primitives = num_vtx / 3;
-               device->DrawIndexedPrimitiveUP(D3DPT_TRIANGLELIST, min_idx, num_idx, num_primitives, indices, D3DFMT_INDEX32, vtx, stride);
+               device->DrawIndexedPrimitiveUP(D3DPT_TRIANGLELIST, min_idx, max_idx + 1, num_primitives, indices, D3DFMT_INDEX32, vtx, stride);
                break;
             };
             case ALLEGRO_PRIM_TRIANGLE_STRIP: {
                num_primitives = num_vtx - 2;
-               device->DrawIndexedPrimitiveUP(D3DPT_TRIANGLESTRIP, min_idx, num_idx, num_primitives, indices, D3DFMT_INDEX32, vtx, stride);
+               device->DrawIndexedPrimitiveUP(D3DPT_TRIANGLESTRIP, min_idx, max_idx + 1, num_primitives, indices, D3DFMT_INDEX32, vtx, stride);
                break;
             };
             case ALLEGRO_PRIM_TRIANGLE_FAN: {
                num_primitives = num_vtx - 2;
-               device->DrawIndexedPrimitiveUP(D3DPT_TRIANGLEFAN, min_idx, num_idx, num_primitives, indices, D3DFMT_INDEX32, vtx, stride);
+               device->DrawIndexedPrimitiveUP(D3DPT_TRIANGLEFAN, min_idx, max_idx + 1, num_primitives, indices, D3DFMT_INDEX32, vtx, stride);
                break;
             };
             case ALLEGRO_PRIM_POINT_LIST: {
