@@ -222,10 +222,18 @@
          case ALLEGRO_PIXEL_FORMAT_RGBA_4444: {                               \
             uint16_t _gp_pixel = *(uint16_t *)(data);                         \
             _AL_MAP_RGBA(color,                                               \
-               _al_rgb_scale_4[(_gp_pixel & 0xF000) >> 12],                      \
-               _al_rgb_scale_4[(_gp_pixel & 0x0F00) >> 8],                       \
-               _al_rgb_scale_4[(_gp_pixel & 0x00F0) >> 4],                       \
-               _al_rgb_scale_4[(_gp_pixel & 0x000F)]);                           \
+               _al_rgb_scale_4[(_gp_pixel & 0xF000) >> 12],                   \
+               _al_rgb_scale_4[(_gp_pixel & 0x0F00) >> 8],                    \
+               _al_rgb_scale_4[(_gp_pixel & 0x00F0) >> 4],                    \
+               _al_rgb_scale_4[(_gp_pixel & 0x000F)]);                        \
+            if (advance)                                                      \
+               data += 2;                                                     \
+            break;                                                            \
+         }                                                                    \
+                                                                              \
+         case ALLEGRO_PIXEL_FORMAT_SINGLE_CHANNEL_8: {                        \
+            uint8_t c = *(uint8_t *)(data);                                   \
+            _AL_MAP_RGBA(color, c, c, c, 255);                                \
             if (advance)                                                      \
                data += 2;                                                     \
             break;                                                            \
@@ -240,13 +248,20 @@
          case ALLEGRO_PIXEL_FORMAT_ANY_24_NO_ALPHA:                           \
          case ALLEGRO_PIXEL_FORMAT_ANY_32_NO_ALPHA:                           \
          case ALLEGRO_PIXEL_FORMAT_ANY_32_WITH_ALPHA:                         \
-            ALLEGRO_ERROR("INLINE_GET got fake _gp_pixel format: %d\n", format); \
+            ALLEGRO_ERROR("INLINE_GET got fake pixel format: %d\n", format); \
+            abort();                                                          \
+            break;                                                            \
+                                                                              \
+         case ALLEGRO_PIXEL_FORMAT_COMPRESSED_RGBA_DXT1:                                 \
+         case ALLEGRO_PIXEL_FORMAT_COMPRESSED_RGBA_DXT3:                                 \
+         case ALLEGRO_PIXEL_FORMAT_COMPRESSED_RGBA_DXT5:                                 \
+            ALLEGRO_ERROR("INLINE_GET got compressed format: %d\n", format); \
             abort();                                                          \
             break;                                                            \
                                                                               \
          case ALLEGRO_NUM_PIXEL_FORMATS:                                      \
          default:                                                             \
-            ALLEGRO_ERROR("INLINE_GET got non _gp_pixel format: %d\n", format); \
+            ALLEGRO_ERROR("INLINE_GET got non pixel format: %d\n", format); \
             abort();                                                          \
             break;                                                            \
       }                                                                       \
@@ -431,6 +446,14 @@
                data += 2;                                                     \
             break;                                                            \
                                                                               \
+         case ALLEGRO_PIXEL_FORMAT_SINGLE_CHANNEL_8: {                        \
+            uint8_t c = color.r;                                              \
+            *(uint8_t *)data = c;                                             \
+            if (advance)                                                      \
+               data += 3;                                                     \
+            break;                                                            \
+         }                                                                    \
+                                                                              \
          case ALLEGRO_PIXEL_FORMAT_ANY:                                       \
          case ALLEGRO_PIXEL_FORMAT_ANY_NO_ALPHA:                              \
          case ALLEGRO_PIXEL_FORMAT_ANY_WITH_ALPHA:                            \
@@ -444,6 +467,13 @@
             abort();                                                          \
             break;                                                            \
                                                                               \
+         case ALLEGRO_PIXEL_FORMAT_COMPRESSED_RGBA_DXT1:                      \
+         case ALLEGRO_PIXEL_FORMAT_COMPRESSED_RGBA_DXT3:                      \
+         case ALLEGRO_PIXEL_FORMAT_COMPRESSED_RGBA_DXT5:                      \
+            ALLEGRO_ERROR("INLINE_PUT got compressed format: %d\n", format); \
+            abort();                                                          \
+            break;                                                            \
+                                                                              \
          case ALLEGRO_NUM_PIXEL_FORMATS:                                      \
             ALLEGRO_ERROR("INLINE_PUT got non _pp_pixel format: %d\n", format); \
             abort();                                                          \
@@ -457,11 +487,13 @@ AL_ARRAY(int, _al_rgb_scale_5);
 AL_ARRAY(int, _al_rgb_scale_6);
 AL_ARRAY(float, _al_u8_to_float);
 
-void _al_init_pixels(void);
-bool _al_pixel_format_has_alpha(int format);
-bool _al_pixel_format_is_real(int format);
-int _al_get_real_pixel_format(ALLEGRO_DISPLAY *display, int format);
-char const *_al_pixel_format_name(ALLEGRO_PIXEL_FORMAT format);
+AL_FUNC(void, _al_init_pixels, (void));
+AL_FUNC(bool, _al_pixel_format_has_alpha, (int format));
+AL_FUNC(bool, _al_pixel_format_is_real, (int format));
+AL_FUNC(bool, _al_pixel_format_is_video_only, (int format));
+AL_FUNC(bool, _al_pixel_format_is_compressed, (int format));
+AL_FUNC(int, _al_get_real_pixel_format, (ALLEGRO_DISPLAY *display, int format));
+AL_FUNC(char const*, _al_pixel_format_name, (ALLEGRO_PIXEL_FORMAT format));
 
 
 #ifdef __cplusplus
