@@ -1,5 +1,5 @@
 /*         ______   ___    ___ 
- *        /\  _  \ /\_ \  /\_ \ 
+ *        /\  _  \ /\_ \  /\_ \
  *        \ \ \L\ \\//\ \ \//\ \      __     __   _ __   ___ 
  *         \ \  __ \ \ \ \  \ \ \   /'__`\ /'_ `\/\`'__\/ __`\
  *          \ \ \/\ \ \_\ \_ \_\ \_/\  __//\ \L\ \ \ \//\ \L\ \
@@ -26,9 +26,8 @@
 #include "allegro5/internal/aintern.h"
 #include "allegro5/internal/aintern_bitmap.h"
 #include "allegro5/internal/aintern_display.h"
+#include "allegro5/internal/aintern_shader.h"
 #include "allegro5/internal/aintern_system.h"
-#include "allegro5/internal/aintern_display.h"
-#include "allegro5/internal/aintern_bitmap.h"
 
 
 ALLEGRO_DEBUG_CHANNEL("display")
@@ -41,7 +40,8 @@ ALLEGRO_DISPLAY *al_create_display(int w, int h)
    ALLEGRO_SYSTEM *system;
    ALLEGRO_DISPLAY_INTERFACE *driver;
    ALLEGRO_DISPLAY *display;
-   ALLEGRO_TRANSFORM identity;
+   ALLEGRO_EXTRA_DISPLAY_SETTINGS *settings;
+   int flags;
 
    system = al_get_system_driver();
    driver = system->vt->get_display_driver();
@@ -58,28 +58,57 @@ ALLEGRO_DISPLAY *al_create_display(int w, int h)
 
    ASSERT(display->vt);
 
+   settings = &display->extra_settings;
+   flags = settings->required | settings->suggested;
+   if (!(flags & (1 << ALLEGRO_AUTO_CONVERT_BITMAPS))) {
+      settings->settings[ALLEGRO_AUTO_CONVERT_BITMAPS] = 1;
+   }
+
+   display->min_w = 0;
+   display->min_h = 0;
+   display->max_w = 0;
+   display->max_h = 0;
+
    display->vertex_cache = 0;
    display->num_cache_vertices = 0;
    display->cache_enabled = false;
    display->vertex_cache_size = 0;
    display->cache_texture = 0;
-   
-   display->display_invalidated = 0;
-   
+   al_identity_transform(&display->projview_transform);
+
+   display->default_shader = NULL;
+
+   _al_vector_init(&display->display_invalidated_callbacks, sizeof(void *));
+   _al_vector_init(&display->display_validated_callbacks, sizeof(void *));
+
+   display->render_state.write_mask = ALLEGRO_MASK_RGBA | ALLEGRO_MASK_DEPTH;
+   display->render_state.depth_test = false;
+   display->render_state.depth_function = ALLEGRO_RENDER_LESS;
+   display->render_state.alpha_test = false;
+   display->render_state.alpha_function = ALLEGRO_RENDER_ALWAYS;
+   display->render_state.alpha_test_value = 0;
+
    _al_vector_init(&display->bitmaps, sizeof(ALLEGRO_BITMAP*));
 
-   if (display->extra_settings.settings[ALLEGRO_COMPATIBLE_DISPLAY])
+   if (settings->settings[ALLEGRO_COMPATIBLE_DISPLAY]) {
       al_set_target_bitmap(al_get_backbuffer(display));
+   }
    else {
       ALLEGRO_DEBUG("ALLEGRO_COMPATIBLE_DISPLAY not set\n");
       _al_set_current_display_only(display);
    }
 
-   al_identity_transform(&identity);
-   al_use_transform(&identity);
+   if (display->flags & ALLEGRO_PROGRAMMABLE_PIPELINE) {
+      display->default_shader = _al_create_default_shader(display->flags);
+      if (!display->default_shader) {
+         al_destroy_display(display);
+         return NULL;
+      }
+      al_use_shader(display->default_shader);
+   }
 
    /* Clear the screen */
-   if (display->extra_settings.settings[ALLEGRO_COMPATIBLE_DISPLAY]) {
+   if (settings->settings[ALLEGRO_COMPATIBLE_DISPLAY]) {
       al_clear_to_color(al_map_rgb(0, 0, 0));
 
       /* TODO:
@@ -92,7 +121,17 @@ ALLEGRO_DISPLAY *al_create_display(int w, int h)
 #endif
    }
 
-   al_set_window_title(display, al_get_app_name());
+   // TODO: Remove this once all the implementations set the title correctly themselves.
+   // Issue #544.
+   al_set_window_title(display, al_get_new_window_title());
+   
+   if (settings->settings[ALLEGRO_AUTO_CONVERT_BITMAPS]) {
+      /* We convert video bitmaps to memory bitmaps when the display is
+       * destroyed, so seems only fair to re-convertt hem when the
+       * display is re-created again.
+       */
+      al_convert_memory_bitmaps();
+   }
 
    return display;
 }
@@ -104,10 +143,16 @@ ALLEGRO_DISPLAY *al_create_display(int w, int h)
 void al_destroy_display(ALLEGRO_DISPLAY *display)
 {
    if (display) {
+      /* This causes warnings and potential errors on Android because
+       * it clears the context and Android needs this thread to have
+       * the context bound in its destroy function and to destroy the
+       * shader. Just skip this part on Android.
+       */
+#ifndef ALLEGRO_ANDROID
       ALLEGRO_BITMAP *bmp;
 
       bmp = al_get_target_bitmap();
-      if (bmp && bmp->display == display)
+      if (bmp && _al_get_bitmap_display(bmp) == display)
          al_set_target_bitmap(NULL);
 
       /* This can happen if we have a current display, but the target bitmap
@@ -115,6 +160,10 @@ void al_destroy_display(ALLEGRO_DISPLAY *display)
        */
       if (display == al_get_current_display())
          _al_set_current_display_only(NULL);
+#endif
+
+      al_destroy_shader(display->default_shader);
+      display->default_shader = NULL;
 
       ASSERT(display->vt);
       display->vt->destroy_display(display);
@@ -267,6 +316,16 @@ int al_get_display_flags(ALLEGRO_DISPLAY *display)
 }
 
 
+/* Function: al_get_display_orientation
+ */
+int al_get_display_orientation(ALLEGRO_DISPLAY* display)
+{
+   if (display && display->vt->get_orientation)
+      return display->vt->get_orientation(display);
+   else
+      return ALLEGRO_DISPLAY_ORIENTATION_UNKNOWN;
+}
+
 
 /* Function: al_wait_for_vsync
  */
@@ -313,20 +372,6 @@ void al_set_display_icons(ALLEGRO_DISPLAY *display,
    }
 }
 
-
-
-/* Destroys all bitmaps created for this display.
- */
-void _al_destroy_display_bitmaps(ALLEGRO_DISPLAY *d)
-{
-   while (_al_vector_size(&d->bitmaps) > 0) {
-      ALLEGRO_BITMAP **bptr = _al_vector_ref_back(&d->bitmaps);
-      ALLEGRO_BITMAP *b = *bptr;
-      al_destroy_bitmap(b);
-   }
-}
-
-
 /* Function: al_set_window_position
  */
 void al_set_window_position(ALLEGRO_DISPLAY *display, int x, int y)
@@ -359,6 +404,65 @@ void al_get_window_position(ALLEGRO_DISPLAY *display, int *x, int *y)
 }
 
 
+/* Function: al_set_window_constraints
+ */
+bool al_set_window_constraints(ALLEGRO_DISPLAY *display,
+   int min_w, int min_h, int max_w, int max_h)
+{
+   ASSERT(display);
+
+  /* Perform some basic checks. */
+   if (min_w < 0 || min_h < 0 || max_w < 0 || max_h < 0) {
+      return false;
+   }
+   if (min_w > 0 && max_w > 0 && max_w < min_w) {
+      return false;
+   }
+   if (min_h > 0 && max_h > 0 && max_h < min_h) {
+      return false;
+   }
+
+   /* Cannot constrain when fullscreen. */
+   if (display->flags & ALLEGRO_FULLSCREEN) {
+      return false;
+   }
+
+   /* Cannot constrain if not resizable. */
+   if (!(display->flags & ALLEGRO_RESIZABLE)) {
+      return false;
+   }
+
+   if (display && display->vt && display->vt->set_window_constraints) {
+      return display->vt->set_window_constraints(display, min_w, min_h,
+         max_w, max_h);
+   }
+   else {
+      return false;
+   }
+}
+
+
+/* Function: al_get_window_constraints
+ */
+bool al_get_window_constraints(ALLEGRO_DISPLAY *display,
+   int *min_w, int *min_h, int *max_w, int *max_h)
+{
+   ASSERT(display);
+   ASSERT(min_w);
+   ASSERT(min_h);
+   ASSERT(max_w);
+   ASSERT(max_h);
+
+   if (display && display->vt && display->vt->get_window_constraints) {
+      return display->vt->get_window_constraints(display, min_w, min_h,
+         max_w, max_h);
+   }
+   else {
+      return false;
+   }
+}
+
+
 /* Function: al_set_display_flag
  */
 bool al_set_display_flag(ALLEGRO_DISPLAY *display, int flag, bool onoff)
@@ -372,16 +476,6 @@ bool al_set_display_flag(ALLEGRO_DISPLAY *display, int flag, bool onoff)
 }
 
 
-/* Function: al_toggle_display_flag
- *  This is deprecated.
- *  On the 5.0 branch this needs to be a function for backwards compatibility.
- */
-bool al_toggle_display_flag(ALLEGRO_DISPLAY *display, int flag, bool onoff)
-{
-   return al_set_display_flag(display, flag, onoff);
-}
-
-
 /* Function: al_set_window_title
  */
 void al_set_window_title(ALLEGRO_DISPLAY *display, const char *title)
@@ -446,9 +540,92 @@ bool al_is_bitmap_drawing_held(void)
       return false;
 }
 
-void _al_set_display_invalidated_callback(ALLEGRO_DISPLAY* display, void (*display_invalidated)(ALLEGRO_DISPLAY*))
+void _al_add_display_invalidated_callback(ALLEGRO_DISPLAY* display, void (*display_invalidated)(ALLEGRO_DISPLAY*))
 {
-   display->display_invalidated = display_invalidated;
+   if (_al_vector_find(&display->display_invalidated_callbacks, display_invalidated) >= 0) {
+      return;
+   }
+   else {
+      void (**callback)(ALLEGRO_DISPLAY *) = _al_vector_alloc_back(&display->display_invalidated_callbacks);
+      *callback = display_invalidated;
+   }
+}
+
+void _al_add_display_validated_callback(ALLEGRO_DISPLAY* display, void (*display_validated)(ALLEGRO_DISPLAY*))
+{
+   if (_al_vector_find(&display->display_validated_callbacks, display_validated) >= 0) {
+      return;
+   }
+   else {
+      void (**callback)(ALLEGRO_DISPLAY *) = _al_vector_alloc_back(&display->display_validated_callbacks);
+      *callback = display_validated;
+   }
+}
+
+void _al_remove_display_invalidated_callback(ALLEGRO_DISPLAY *display, void (*callback)(ALLEGRO_DISPLAY *))
+{
+   _al_vector_find_and_delete(&display->display_invalidated_callbacks, &callback);
+}
+
+void _al_remove_display_validated_callback(ALLEGRO_DISPLAY *display, void (*callback)(ALLEGRO_DISPLAY *))
+{
+   _al_vector_find_and_delete(&display->display_validated_callbacks, &callback);
+}
+
+/* Function: al_acknowledge_drawing_halt
+ */
+void al_acknowledge_drawing_halt(ALLEGRO_DISPLAY *display)
+{
+   if (display->vt->acknowledge_drawing_halt) {
+      display->vt->acknowledge_drawing_halt(display);
+   }
+}
+
+/* Function: al_acknowledge_drawing_resume
+ */
+void al_acknowledge_drawing_resume(ALLEGRO_DISPLAY *display)
+{
+   if (display->vt->acknowledge_drawing_resume) {
+      display->vt->acknowledge_drawing_resume(display);
+   }
+}
+
+/* Function: al_set_render_state
+ */
+void al_set_render_state(ALLEGRO_RENDER_STATE state, int value)
+{
+   ALLEGRO_DISPLAY *display = al_get_current_display();
+
+   if (!display)
+      return;
+
+   switch (state) {
+      case ALLEGRO_ALPHA_TEST:
+         display->render_state.alpha_test = value;
+         break;
+      case ALLEGRO_WRITE_MASK:
+         display->render_state.write_mask = value;
+         break;
+      case ALLEGRO_DEPTH_TEST:
+         display->render_state.depth_test = value;
+         break;
+      case ALLEGRO_DEPTH_FUNCTION:
+         display->render_state.depth_function = value;
+         break;
+      case ALLEGRO_ALPHA_FUNCTION:
+         display->render_state.alpha_function = value;
+         break;
+      case ALLEGRO_ALPHA_TEST_VALUE:
+         display->render_state.alpha_test_value = value;
+         break;
+      default:
+         ALLEGRO_WARN("unknown state to change: %d\n", state);
+         break;
+   }
+
+   if (display->vt && display->vt->update_render_state) {
+      display->vt->update_render_state(display);
+   }
 }
 
 /* vim: set sts=3 sw=3 et: */
