@@ -37,7 +37,7 @@
 
 #if defined ALLEGRO_MACOSX
 #include <OpenGL/glu.h>
-#elif !defined ALLEGRO_GP2XWIZ && !defined ALLEGRO_IPHONE
+#elif !defined ALLEGRO_CFG_OPENGLES
 #include <GL/glu.h>
 #endif
 
@@ -58,7 +58,7 @@ ALLEGRO_DEBUG_CHANNEL("opengl")
     */
    #if defined ALLEGRO_GLXGETPROCADDRESSARB
       #define alXGetProcAddress glXGetProcAddressARB
-   #elif defined ALLEGRO_GP2XWIZ
+   #elif defined ALLEGRO_RASPBERRYPI
       #define alXGetProcAddress eglGetProcAddress
    #else
       #define alXGetProcAddress glXGetProcAddress
@@ -91,7 +91,7 @@ ALLEGRO_DEBUG_CHANNEL("opengl")
 
 
 
-#ifndef ALLEGRO_IPHONE
+#if !defined ALLEGRO_CFG_OPENGLES
 static uint32_t parse_opengl_version(const char *s)
 {
    char *p = (char *) s;
@@ -125,23 +125,19 @@ static uint32_t parse_opengl_version(const char *s)
 /* Reads version info out of glGetString(GL_VERSION) */
 static uint32_t _al_ogl_version(void)
 {
-#ifndef ALLEGRO_IPHONE
-   ALLEGRO_CONFIG *cfg;
+#if !defined ALLEGRO_CFG_OPENGLES
    const char *str;
 
-   cfg = al_get_system_config();
-   if (cfg) {
-      char const *value = al_get_config_value(cfg,
-	 "opengl", "force_opengl_version");
-      if (value) {
-         uint32_t v = parse_opengl_version(value);
-         ALLEGRO_INFO("OpenGL version forced to %d.%d.%d.%d.\n",
-            (v >> 24) & 0xff,
-            (v >> 16) & 0xff,
-            (v >> 8) & 0xff,
-            (v & 0xff));
-         return v;
-      }
+   char const *value = al_get_config_value(al_get_system_config(), "opengl",
+      "force_opengl_version");
+   if (value) {
+      uint32_t v = parse_opengl_version(value);
+      ALLEGRO_INFO("OpenGL version forced to %d.%d.%d.%d.\n",
+         (v >> 24) & 0xff,
+         (v >> 16) & 0xff,
+         (v >> 8) & 0xff,
+         (v & 0xff));
+      return v;
    }
 
    str = (const char *)glGetString(GL_VERSION);
@@ -203,6 +199,7 @@ static void print_extensions(char const *extension)
 
 
 
+#if !defined ALLEGRO_CFG_OPENGLES
 /* Print all extensions the OpenGL 3.0 way. */ 
 static void print_extensions_3_0(void)
 {
@@ -215,6 +212,7 @@ static void print_extensions_3_0(void)
       ALLEGRO_DEBUG("%s\n", name);
    }
 }
+#endif
 
 
 
@@ -240,7 +238,7 @@ uint32_t al_get_opengl_version(void)
  */
 int al_get_opengl_variant(void)
 {
-#if defined ALLEGRO_IPHONE || defined ALLEGRO_GP2XWIZ
+#if defined ALLEGRO_CFG_OPENGLES
    return ALLEGRO_OPENGL_ES;
 #else
    return ALLEGRO_DESKTOP_OPENGL;
@@ -275,6 +273,36 @@ static ALLEGRO_OGL_EXT_API *create_extension_api_table(void)
 
 
 
+typedef void (*VOID_FPTR)(void);
+/* GCC 4.8.2 and possibly others are really slow at optimizing the 100's of the
+ * if statements in the load_extensions function below, so we extract them to
+ * this function.
+ */
+static VOID_FPTR load_extension(const char* name)
+{
+   VOID_FPTR fptr = NULL;
+#ifdef ALLEGRO_WINDOWS
+   fptr = (VOID_FPTR)wglGetProcAddress(name);
+#elif defined ALLEGRO_UNIX
+   fptr = (VOID_FPTR)alXGetProcAddress((const GLubyte*)name);
+#elif defined ALLEGRO_MACOSX
+   CFStringRef cfstr = CFStringCreateWithCStringNoCopy(NULL, name,
+      kCFStringEncodingUTF8, kCFAllocatorNull);
+   if (cfstr) {
+      fptr = (VOID_FPTR)CFBundleGetFunctionPointerForName(opengl_bundle_ref, cfstr);
+      CFRelease(cfstr);
+   }
+#elif defined ALLEGRO_SDL
+   fptr = SDL_GL_GetProcAddress(name);
+#endif
+   if (fptr) {
+      ALLEGRO_DEBUG("%s successfully loaded (%p)\n", name, fptr);
+   }
+   return fptr;
+}
+
+
+
 /* Load the extension API addresses into the table.
  * Should only be done on context creation.
  */
@@ -294,17 +322,15 @@ static void load_extensions(ALLEGRO_OGL_EXT_API *ext)
 
 #ifdef ALLEGRO_WINDOWS
 
-   #define AGL_API(type, name, args)                                 \
-      ext->name = (_ALLEGRO_gl##name##_t)wglGetProcAddress("gl" #name); \
-      if (ext->name) { ALLEGRO_DEBUG("gl" #name " successfully loaded\n"); }
+   #define AGL_API(type, name, args)                                           \
+      ext->name = (_ALLEGRO_gl##name##_t)load_extension("gl" #name);
 
       #include "allegro5/opengl/GLext/gl_ext_api.h"
 
    #undef AGL_API
 
-   #define AGL_API(type, name, args)                                  \
-      ext->name = (_ALLEGRO_wgl##name##_t)wglGetProcAddress("wgl" #name); \
-      if (ext->name) { ALLEGRO_DEBUG("wgl" #name " successfully loaded\n"); }
+   #define AGL_API(type, name, args)                                           \
+      ext->name = (_ALLEGRO_wgl##name##_t)load_extension("wgl" #name);
 
       #include "allegro5/opengl/GLext/wgl_ext_api.h"
 
@@ -312,17 +338,15 @@ static void load_extensions(ALLEGRO_OGL_EXT_API *ext)
 
 #elif defined ALLEGRO_UNIX
 
-   #define AGL_API(type, name, args)                                               \
-      ext->name = (_ALLEGRO_gl##name##_t)alXGetProcAddress((const GLubyte*)"gl" #name);  \
-      if (ext->name) { ALLEGRO_DEBUG("gl" #name " successfully loaded (%p)\n", ext->name); }
+   #define AGL_API(type, name, args)                                           \
+      ext->name = (_ALLEGRO_gl##name##_t)load_extension("gl" #name);
 
       #include "allegro5/opengl/GLext/gl_ext_api.h"
 
    #undef AGL_API
 
-   #define AGL_API(type, name, args)                                               \
-      ext->name = (_ALLEGRO_glX##name##_t)alXGetProcAddress((const GLubyte*)"glX" #name); \
-      if (ext->name) { ALLEGRO_DEBUG("glX" #name " successfully loaded\n"); }
+   #define AGL_API(type, name, args)                                           \
+      ext->name = (_ALLEGRO_glX##name##_t)load_extension("glX" #name);
 
       #include "allegro5/opengl/GLext/glx_ext_api.h"
 
@@ -330,9 +354,17 @@ static void load_extensions(ALLEGRO_OGL_EXT_API *ext)
 
 #elif defined ALLEGRO_MACOSX
 
-#define AGL_API(type, name, args)                                                                 \
-      ext->name = (_ALLEGRO_gl##name##_t)CFBundleGetFunctionPointerForName(opengl_bundle_ref, CFSTR("gl" # name)); \
-      if (ext->name) { ALLEGRO_DEBUG("gl" #name " successfully loaded\n"); }
+#define AGL_API(type, name, args)                                              \
+      ext->name = (_ALLEGRO_gl##name##_t)load_extension("gl" # name);
+
+      #include "allegro5/opengl/GLext/gl_ext_api.h"
+
+   #undef AGL_API
+
+#elif defined ALLEGRO_SDL
+
+#define AGL_API(type, name, args)                                              \
+      ext->name = (_ALLEGRO_gl##name##_t)load_extension("gl" # name);
 
       #include "allegro5/opengl/GLext/gl_ext_api.h"
 
@@ -429,14 +461,12 @@ static bool _ogl_is_extension_supported(const char *extension,
 {
    int ret = 0;
    GLubyte const *ext_str;
+#if !defined ALLEGRO_CFG_OPENGLES
    int v = al_get_opengl_version();
+#endif
    (void)disp;
 
-#if defined ALLEGRO_GP2XWIZ
-   return false;
-#endif
-
-#ifndef ALLEGRO_IPHONE
+#if !defined ALLEGRO_CFG_OPENGLES
    if (disp->flags & ALLEGRO_OPENGL_3_0 || v >= _ALLEGRO_OPENGL_VERSION_3_0) {
       int i;
       GLint ext_cnt;
@@ -500,7 +530,6 @@ static bool _ogl_is_extension_supported(const char *extension,
 static bool _ogl_is_extension_with_version_supported(
    const char *extension, ALLEGRO_DISPLAY *disp, uint32_t ver)
 {
-   ALLEGRO_CONFIG *cfg;
    char const *value;
 
   /* For testing purposes, any OpenGL extension can be disable in
@@ -511,15 +540,12 @@ static bool _ogl_is_extension_with_version_supported(
     * GL_EXT_framebuffer_object=0
     * 
     */
-   cfg = al_get_system_config();
-   if (cfg) {
-      value = al_get_config_value(cfg,
-         "opengl_disabled_extensions", extension);
-      if (value) {
-         ALLEGRO_WARN("%s found in [opengl_disabled_extensions].\n",
-            extension);
-         return false;
-      }
+   value = al_get_config_value(al_get_system_config(),
+      "opengl_disabled_extensions", extension);
+   if (value) {
+      ALLEGRO_WARN("%s found in [opengl_disabled_extensions].\n",
+         extension);
+      return false;
    }
 
    /* If the extension is included in the OpenGL version, there is no
@@ -592,7 +618,7 @@ void *al_get_opengl_proc_address(const char *name)
        * address. Unfortunately glXGetProcAddress is an extension
        * and may not be available on all platforms
        */
-#ifdef ALLEGRO_GP2XWIZ
+#if defined ALLEGRO_RASPBERRYPI
       symbol = alXGetProcAddress(name);
 #else
       symbol = alXGetProcAddress((const GLubyte *)name);
@@ -659,6 +685,9 @@ static void fill_in_info_struct(const GLubyte *rendereru, OPENGL_INFO *info)
    else if (strstr(renderer, "Mesa DRI R200")) {
       info->is_ati_r200_chip = 1;
    }
+   else if (strstr(renderer, "Intel HD Graphics 3000 OpenGL Engine")) {
+      info->is_intel_hd_graphics_3000 = 1;
+   }
 
    if ((strncmp(renderer, "3Dfx/Voodoo3 ", 13) == 0)
        || (strncmp(renderer, "3Dfx/Voodoo2 ", 13) == 0)
@@ -762,6 +791,7 @@ void _al_ogl_manage_extensions(ALLEGRO_DISPLAY *gl_disp)
    load_extensions(ext_api);
    gl_disp->ogl_extras->extension_api = ext_api;
    
+#if !defined ALLEGRO_CFG_OPENGLES
    /* Need that symbol already so can't wait until it is assigned later. */
    glGetStringi = ext_api->GetStringi;
 
@@ -769,6 +799,7 @@ void _al_ogl_manage_extensions(ALLEGRO_DISPLAY *gl_disp)
       ALLEGRO_DEBUG("OpenGL Extensions:\n");
       print_extensions_3_0();
    }
+#endif
 
    /* Create the list of supported extensions. */
    ext_list = create_extension_list();
@@ -869,12 +900,18 @@ void _al_ogl_manage_extensions(ALLEGRO_DISPLAY *gl_disp)
          s[ALLEGRO_SUPPORT_SEPARATE_ALPHA] = 1;
 
       s[ALLEGRO_SUPPORT_NPOT_BITMAP] =
-         ext_list->ALLEGRO_GL_ARB_texture_non_power_of_two;
+         ext_list->ALLEGRO_GL_ARB_texture_non_power_of_two ||
+         ext_list->ALLEGRO_GL_OES_texture_npot;
    ALLEGRO_INFO("Use of non-power-of-two textures %s.\n",
       s[ALLEGRO_SUPPORT_NPOT_BITMAP] ? "enabled" : "disabled");
-#ifdef ALLEGRO_IPHONE
-   s[ALLEGRO_CAN_DRAW_INTO_BITMAP] =
-      ext_list->ALLEGRO_GL_OES_framebuffer_object;
+#if defined ALLEGRO_CFG_OPENGLES
+   if (gl_disp->flags & ALLEGRO_PROGRAMMABLE_PIPELINE) {
+      s[ALLEGRO_CAN_DRAW_INTO_BITMAP] = true;
+   }
+   else {
+      s[ALLEGRO_CAN_DRAW_INTO_BITMAP] =
+         ext_list->ALLEGRO_GL_OES_framebuffer_object;
+   }
    ALLEGRO_INFO("Use of FBO to draw to textures %s.\n",
       s[ALLEGRO_CAN_DRAW_INTO_BITMAP] ? "enabled" :
       "disabled");
