@@ -69,7 +69,7 @@ ALLEGRO_SHADER *al_create_shader(ALLEGRO_SHADER_PLATFORM platform)
    if (shader) {
       ASSERT(shader->platform);
       ASSERT(shader->vt);
-      _al_register_destructor(_al_dtor_list, shader,
+      _al_register_destructor(_al_dtor_list, "shader", shader,
          (void (*)(void *))al_destroy_shader);
    }
    else {
@@ -150,19 +150,21 @@ ALLEGRO_SHADER_PLATFORM al_get_shader_platform(ALLEGRO_SHADER *shader)
 bool al_use_shader(ALLEGRO_SHADER *shader)
 {
    ALLEGRO_BITMAP *bmp = al_get_target_bitmap();
+   ALLEGRO_DISPLAY *disp;
 
    if (!bmp) {
       ALLEGRO_WARN("No current target bitmap.\n");
       return false;
    }
-   if (bmp->flags & ALLEGRO_MEMORY_BITMAP) {
+   if (al_get_bitmap_flags(bmp) & ALLEGRO_MEMORY_BITMAP) {
       ALLEGRO_WARN("Target bitmap is memory bitmap.\n");
       return false;
    }
-   ASSERT(bmp->display);
+   disp = _al_get_bitmap_display(bmp);
+   ASSERT(disp);
 
    if (shader) {
-      if (shader->vt->use_shader(shader, bmp->display, true)) {
+      if (shader->vt->use_shader(shader, disp, true)) {
          _al_set_bitmap_shader_field(bmp, shader);
          ALLEGRO_DEBUG("use_shader succeeded\n");
          return true;
@@ -170,21 +172,21 @@ bool al_use_shader(ALLEGRO_SHADER *shader)
       else {
          _al_set_bitmap_shader_field(bmp, NULL);
          ALLEGRO_ERROR("use_shader failed\n");
-         if (bmp->display->default_shader) {
-            bmp->display->default_shader->vt->use_shader(
-               bmp->display->default_shader, bmp->display, true);
+         if (disp->default_shader) {
+            disp->default_shader->vt->use_shader(
+               disp->default_shader, disp, true);
          }
          return false;
       }
    }
    else {
       if (bmp->shader) {
-         bmp->shader->vt->unuse_shader(bmp->shader, bmp->display);
+         bmp->shader->vt->unuse_shader(bmp->shader, disp);
          _al_set_bitmap_shader_field(bmp, NULL);
       }
-      if (bmp->display->default_shader) {
-         bmp->display->default_shader->vt->use_shader(
-            bmp->display->default_shader, bmp->display, true);
+      if (disp->default_shader) {
+         disp->default_shader->vt->use_shader(
+            disp->default_shader, disp, true);
       }
       return true;
    }
@@ -379,6 +381,7 @@ bool al_set_shader_bool(const char *name, bool b)
 char const *al_get_default_shader_source(ALLEGRO_SHADER_PLATFORM platform,
    ALLEGRO_SHADER_TYPE type)
 {
+   (void)type;
    switch (resolve_platform(platform)) {
       case ALLEGRO_SHADER_GLSL:
 #ifdef ALLEGRO_CFG_SHADER_GLSL
@@ -447,6 +450,7 @@ ALLEGRO_SHADER *_al_create_default_shader(int display_flags)
 {
    ALLEGRO_SHADER_PLATFORM platform = ALLEGRO_SHADER_AUTO;
    ALLEGRO_SHADER *shader;
+   (void)display_flags;
 
    if (false) {
    }
