@@ -20,6 +20,10 @@
 
 #include "allegro5/allegro.h"
 #include "allegro5/internal/aintern.h"
+#ifdef ALLEGRO_CFG_SHADER_GLSL
+#include "allegro5/allegro_opengl.h"
+#include "allegro5/internal/aintern_opengl.h"
+#endif
 #include ALLEGRO_INTERNAL_HEADER
 #include "allegro5/internal/aintern_bitmap.h"
 #include "allegro5/internal/aintern_debug.h"
@@ -35,6 +39,7 @@
 ALLEGRO_DEBUG_CHANNEL("system")
 
 static ALLEGRO_SYSTEM *active_sysdrv = NULL;
+static ALLEGRO_CONFIG *sys_config = NULL;
 
 _AL_VECTOR _al_system_interfaces;
 static _AL_VECTOR _user_system_interfaces = _AL_VECTOR_INITIALIZER(ALLEGRO_SYSTEM_INTERFACE *);
@@ -70,20 +75,19 @@ static ALLEGRO_SYSTEM *find_system(_AL_VECTOR *vector)
 static void shutdown_system_driver(void)
 {
    if (active_sysdrv) {
-      ALLEGRO_CONFIG *temp = active_sysdrv->config;
       if (active_sysdrv->user_exe_path)
          al_destroy_path(active_sysdrv->user_exe_path);
       if (active_sysdrv->vt && active_sysdrv->vt->shutdown_system)
          active_sysdrv->vt->shutdown_system();
       active_sysdrv = NULL;
-      /* active_sysdrv is not accessible here so we copied it */
-      al_destroy_config(temp);
  
       while (!_al_vector_is_empty(&_al_system_interfaces))
          _al_vector_delete_at(&_al_system_interfaces, _al_vector_size(&_al_system_interfaces)-1);
       _al_vector_free(&_al_system_interfaces);
       _al_vector_init(&_al_system_interfaces, sizeof(ALLEGRO_SYSTEM_INTERFACE *));
    }
+   al_destroy_config(sys_config);
+   sys_config = NULL;
 }
 
 
@@ -102,6 +106,10 @@ static ALLEGRO_PATH *early_get_exename_path(void)
    return _al_iphone_get_path(ALLEGRO_EXENAME_PATH);
 #elif defined(ALLEGRO_UNIX)
    return _al_unix_get_path(ALLEGRO_EXENAME_PATH);
+#elif defined(ALLEGRO_ANDROID)
+   return _al_android_get_path(ALLEGRO_EXENAME_PATH);
+#elif defined(ALLEGRO_SDL)
+   return al_create_path_for_directory(SDL_GetBasePath());
 #else
    #error early_get_exename_path not implemented
 #endif
@@ -111,33 +119,28 @@ static ALLEGRO_PATH *early_get_exename_path(void)
 
 static void read_allegro_cfg(void)
 {
-   /* We cannot use any logging in this function as it would cause the
-    * logging system to be initialised before all the relevant config files
-    * have been read in.
-    */
-
    /* We assume that the stdio file interface is in effect. */
 
    ALLEGRO_PATH *path;
    ALLEGRO_CONFIG *temp;
 
-   active_sysdrv->config = NULL;
+   if (!sys_config)
+      sys_config = al_create_config();
 
-#if defined(ALLEGRO_UNIX) && !defined(ALLEGRO_GP2XWIZ) && !defined(ALLEGRO_IPHONE)
-   active_sysdrv->config = al_load_config_file("/etc/allegro5rc");
+#if defined(ALLEGRO_UNIX) && !defined(ALLEGRO_IPHONE)
+   temp = al_load_config_file("/etc/allegro5rc");
+   if (temp) {
+      al_merge_config_into(sys_config, temp);
+      al_destroy_config(temp);
+   }
 
    path = _al_unix_get_path(ALLEGRO_USER_HOME_PATH);
    if (path) {
       al_set_path_filename(path, "allegro5rc");
       temp = al_load_config_file(al_path_cstr(path, '/'));
       if (temp) {
-         if (active_sysdrv->config) {
-            al_merge_config_into(active_sysdrv->config, temp);
-            al_destroy_config(temp);
-         }
-         else {
-            active_sysdrv->config = temp;
-         }
+         al_merge_config_into(sys_config, temp);
+         al_destroy_config(temp);
       }
       al_destroy_path(path);
    }
@@ -148,22 +151,13 @@ static void read_allegro_cfg(void)
       al_set_path_filename(path, "allegro5.cfg");
       temp = al_load_config_file(al_path_cstr(path, ALLEGRO_NATIVE_PATH_SEP));
       if (temp) {
-         if (active_sysdrv->config) {
-            al_merge_config_into(active_sysdrv->config, temp);
-            al_destroy_config(temp);
-         }
-         else {
-            active_sysdrv->config = temp;
-         }
+         al_merge_config_into(sys_config, temp);
+         al_destroy_config(temp);
       }
       al_destroy_path(path);
    }
-
-   /* Always have a configuration available whether or not a config file
-    * exists.
-    */
-   if (!active_sysdrv->config)
-      active_sysdrv->config = al_create_config();
+   /* Reconfigure logging in case something changed. */
+   _al_configure_logging();
 }
 
 
@@ -172,21 +166,31 @@ static void read_allegro_cfg(void)
  * Let a = xa.ya.za.*
  * Let b = xb.yb.zb.*
  *
- * When ya is odd, a is compatible with b if xa.ya.za = xb.yb.zb.
- * When ya is even, a is compatible with b if xa.ya = xb.yb.
+ * When a has the unstable bit, a is compatible with b if xa.ya.za = xb.yb.zb.
+ * Otherwise, a is compatible with b if xa.ya = xb.yb.
  *
  * Otherwise a and b are incompatible.
  */
 static bool compatible_versions(int a, int b)
 {
-   if ((a & 0xffff0000) != (b & 0xffff0000)) {
+   int a_unstable = a & (1 << 31);
+
+   int a_major = (a & 0x7f000000) >> 24;
+   int a_sub   = (a & 0x00ff0000) >> 16;
+   int a_wip   = (a & 0x0000ff00) >> 8;
+
+   int b_major = (b & 0x7f000000) >> 24;
+   int b_sub   = (b & 0x00ff0000) >> 16;
+   int b_wip   = (b & 0x0000ff00) >> 8;
+
+   if (a_major != b_major) {
       return false;
    }
-   if (((a & 0x00ff0000) >> 16) & 1) {
-      
-      if ((a & 0x0000ff00) != (b & 0x0000ff00)) {
-         return false;
-      }
+   if (a_unstable && a_wip != b_wip) {
+      return false;
+   }
+   if (a_sub != b_sub) {
+      return false;
    }
    return true;
 }
@@ -200,7 +204,7 @@ bool al_install_system(int version, int (*atexit_ptr)(void (*)(void)))
    ALLEGRO_SYSTEM bootstrap;
    ALLEGRO_SYSTEM *real_system;
    int library_version = al_get_allegro_version();
-   
+
    if (active_sysdrv) {
       return true;
    }
@@ -216,11 +220,7 @@ bool al_install_system(int version, int (*atexit_ptr)(void (*)(void)))
 
    _al_vector_init(&_al_system_interfaces, sizeof(ALLEGRO_SYSTEM_INTERFACE *));
 
-   /* We want active_sysdrv->config to be available as soon as
-    * possible - for example what if a system driver need to read
-    * settings out of there. Hence we use a dummy ALLEGRO_SYSTEM
-    * here to load the initial config.
-    */
+   /* Set up a bootstrap system so the calls expecting it don't freak out */
    memset(&bootstrap, 0, sizeof(bootstrap));
    active_sysdrv = &bootstrap;
    read_allegro_cfg();
@@ -247,7 +247,7 @@ bool al_install_system(int version, int (*atexit_ptr)(void (*)(void)))
    }
    
    active_sysdrv = real_system;
-   active_sysdrv->config = bootstrap.config;
+   active_sysdrv->mouse_wheel_precision = 1;
 
    ALLEGRO_INFO("Allegro version: %s\n", ALLEGRO_VERSION_STR);
 
@@ -264,9 +264,18 @@ bool al_install_system(int version, int (*atexit_ptr)(void (*)(void)))
    _al_init_pixels();
 
    _al_init_iio_table();
+   
+   _al_init_convert_bitmap_list();
 
    _al_init_timers();
 
+#ifdef ALLEGRO_CFG_SHADER_GLSL
+   _al_glsl_init_shaders();
+#endif
+
+   if (active_sysdrv->vt->heartbeat_init)
+      active_sysdrv->vt->heartbeat_init();
+
    if (atexit_ptr && atexit_virgin) {
       atexit_ptr(al_uninstall_system);
       atexit_virgin = false;
@@ -282,16 +291,25 @@ bool al_install_system(int version, int (*atexit_ptr)(void (*)(void)))
    return true;
 }
 
-
-
 /* Function: al_uninstall_system
  */
 void al_uninstall_system(void)
 {
+   /* Note: al_uninstall_system may get called multiple times without an
+    * al_install_system in between. For example if the user manually
+    * calls it at the end of the program it is called right again
+    * because it's installed as an atexit function by al_init.
+    */
+
    _al_run_destructors(_al_dtor_list);
    _al_run_exit_funcs();
    _al_shutdown_destructors(_al_dtor_list);
    _al_dtor_list = NULL;
+
+#ifdef ALLEGRO_CFG_SHADER_GLSL
+   _al_glsl_shutdown_shaders();
+#endif
+
    _al_shutdown_logging();
 
    /* shutdown_system_driver is registered as an exit func so we don't need
@@ -325,7 +343,9 @@ ALLEGRO_SYSTEM *al_get_system_driver(void)
  */
 ALLEGRO_CONFIG *al_get_system_config(void)
 {
-   return (active_sysdrv) ? active_sysdrv->config : NULL;
+   if (!sys_config)
+      sys_config = al_create_config();
+   return sys_config;
 }
 
 
