@@ -1,6 +1,6 @@
-/*         ______   ___    ___ 
- *        /\  _  \ /\_ \  /\_ \ 
- *        \ \ \L\ \\//\ \ \//\ \      __     __   _ __   ___ 
+/*         ______   ___    ___
+ *        /\  _  \ /\_ \  /\_ \
+ *        \ \ \L\ \\//\ \ \//\ \      __     __   _ __   ___
  *         \ \  __ \ \ \ \  \ \ \   /'__`\ /'_ `\/\`'__\/ __`\
  *          \ \ \/\ \ \_\ \_ \_\ \_/\  __//\ \L\ \ \ \//\ \L\ \
  *           \ \_\ \_\/\____\/\____\ \____\ \____ \ \_\\ \____/
@@ -15,19 +15,19 @@
  */
 
 /* FIXME:
- * 
+ *
  * There are several ways to get thread local storage:
- * 
+ *
  * 1. pthreads.
  * 2. __thread keyword in gcc.
  * 3. __declspec(thread) in MSVC.
  * 4. TLS API under Windows.
- * 
+ *
  * Since pthreads is available from the system everywhere except in
  * Windows, this is the only case which is problematic. It appears
  * that except for old mingw versions (before gcc 4.2) we can simply
  * use __thread, and for MSVC we can always use __declspec(thread):
- * 
+ *
  * However there also is a WANT_TLS configuration variable which is on
  * by default and forces use of the TLS API instead. At the same time,
  * the implementation using the TLS API in this file does not work
@@ -42,8 +42,13 @@
 #include "allegro5/internal/aintern_display.h"
 #include "allegro5/internal/aintern_file.h"
 #include "allegro5/internal/aintern_fshook.h"
+#include "allegro5/internal/aintern_shader.h"
 #include "allegro5/internal/aintern_tls.h"
 
+#ifdef ALLEGRO_ANDROID
+#include "allegro5/internal/aintern_android.h"
+#endif
+
 #if defined(ALLEGRO_MINGW32) && !defined(ALLEGRO_CFG_DLL_TLS)
    /*
     * MinGW < 4.2.1 doesn't have builtin thread local storage, so we
@@ -89,6 +94,15 @@ typedef struct thread_local_state {
    /* Error code */
    int allegro_errno;
 
+   /* Title to use for a new window/display.
+    * This is a static buffer for API reasons.
+    */
+   char new_window_title[ALLEGRO_NEW_WINDOW_TITLE_MAX_SIZE + 1];
+
+#ifdef ALLEGRO_ANDROID
+   JNIEnv *jnienv;
+#endif
+
    /* Destructor ownership count */
    int dtor_owner_count;
 } thread_local_state;
@@ -98,6 +112,7 @@ typedef struct INTERNAL_STATE {
    thread_local_state tls;
    ALLEGRO_BLENDER stored_blender;
    ALLEGRO_TRANSFORM stored_transform;
+   ALLEGRO_TRANSFORM stored_projection_transform;
    int flags;
 } INTERNAL_STATE;
 
@@ -112,6 +127,7 @@ static void initialize_blender(ALLEGRO_BLENDER *b)
    b->blend_alpha_op = ALLEGRO_ADD;
    b->blend_alpha_source = ALLEGRO_ONE;
    b->blend_alpha_dest = ALLEGRO_INVERSE_ALPHA;
+   b->blend_color = al_map_rgba_f(1.0f, 1.0f, 1.0f, 1.0f);
 }
 
 
@@ -124,11 +140,12 @@ static void initialize_tls_values(thread_local_state *tls)
    tls->new_window_y = INT_MAX;
 
    initialize_blender(&tls->current_blender);
-   tls->new_bitmap_flags = ALLEGRO_VIDEO_BITMAP;
+   tls->new_bitmap_flags = ALLEGRO_CONVERT_BITMAP;
    tls->new_bitmap_format = ALLEGRO_PIXEL_FORMAT_ANY_WITH_ALPHA;
    tls->new_file_interface = &_al_file_interface_stdio;
    tls->fs_interface = &_al_fs_interface_stdio;
-   
+   memset(tls->new_window_title, 0, ALLEGRO_NEW_WINDOW_TITLE_MAX_SIZE + 1);
+
    _al_fill_display_settings(&tls->new_display_settings);
 }
 
@@ -142,7 +159,7 @@ static void initialize_tls_values(thread_local_state *tls)
 
 #if defined(ALLEGRO_CFG_DLL_TLS)
    #include "tls_dll.inc"
-#elif defined(ALLEGRO_MACOSX) || defined(ALLEGRO_IPHONE)
+#elif defined(ALLEGRO_MACOSX) || defined(ALLEGRO_IPHONE) || defined(ALLEGRO_ANDROID) || defined(ALLEGRO_RASPBERRYPI)
    #include "tls_pthread.inc"
 #else
    #include "tls_native.inc"
@@ -170,6 +187,46 @@ ALLEGRO_EXTRA_DISPLAY_SETTINGS *_al_get_new_display_settings(void)
 }
 
 
+/* Function: al_set_new_window_title
+ */
+void al_set_new_window_title(const char *title)
+{
+   thread_local_state *tls;
+   size_t size;
+
+   if ((tls = tls_get()) == NULL)
+      return;
+
+   size = strlen(title);
+
+   if (size > ALLEGRO_NEW_WINDOW_TITLE_MAX_SIZE) {
+      size = ALLEGRO_NEW_WINDOW_TITLE_MAX_SIZE;
+   }
+
+   _al_sane_strncpy(tls->new_window_title, title, size + 1);
+}
+
+
+
+/* Function: al_get_new_window_title
+ */
+const char *al_get_new_window_title(void)
+{
+   thread_local_state *tls;
+
+   /* Return app name in case of error or if not set before. */
+   if ((tls = tls_get()) == NULL)
+      return al_get_app_name();
+
+   if (strlen(tls->new_window_title) < 1)
+      return al_get_app_name();
+
+
+   return (const char *)tls->new_window_title;
+}
+
+
+
 
 /* Function: al_set_new_display_flags
  */
@@ -338,24 +395,52 @@ void al_set_target_bitmap(ALLEGRO_BITMAP *bitmap)
    thread_local_state *tls;
    ALLEGRO_DISPLAY *old_display;
    ALLEGRO_DISPLAY *new_display;
+   ALLEGRO_SHADER *old_shader;
+   ALLEGRO_SHADER *new_shader;
+   bool same_shader;
+   int bitmap_flags = bitmap ? al_get_bitmap_flags(bitmap) : 0;
 
    ASSERT(!al_is_bitmap_drawing_held());
 
+   if (bitmap) {
+      if (bitmap->parent) {
+         bitmap->parent->dirty = true;
+      }
+      else {
+         bitmap->dirty = true;
+      }
+   }
+
    if ((tls = tls_get()) == NULL)
       return;
 
    old_display = tls->current_display;
 
+   if (tls->target_bitmap)
+      old_shader = tls->target_bitmap->shader;
+   else
+      old_shader = NULL;
+
    if (bitmap == NULL) {
       /* Explicitly releasing the current rendering context. */
       new_display = NULL;
+      new_shader = NULL;
    }
-   else if (bitmap->flags & ALLEGRO_MEMORY_BITMAP) {
+   else if (bitmap_flags & ALLEGRO_MEMORY_BITMAP) {
       /* Setting a memory bitmap doesn't change the rendering context. */
       new_display = old_display;
+      new_shader = NULL;
    }
    else {
-      new_display = bitmap->display;
+      new_display = _al_get_bitmap_display(bitmap);
+      new_shader = bitmap->shader;
+   }
+
+   same_shader = (old_shader == new_shader && old_display == new_display);
+
+   /* Unset the old shader if necessary. */
+   if (old_shader && !same_shader) {
+      old_shader->vt->unuse_shader(old_shader, old_display);
    }
 
    /* Change the rendering context if necessary. */
@@ -379,12 +464,21 @@ void al_set_target_bitmap(ALLEGRO_BITMAP *bitmap)
    tls->target_bitmap = bitmap;
 
    if (bitmap &&
-         !(bitmap->flags & ALLEGRO_MEMORY_BITMAP) &&
+         !(bitmap_flags & ALLEGRO_MEMORY_BITMAP) &&
          new_display &&
          new_display->vt &&
-         new_display->vt->set_target_bitmap) {
+         new_display->vt->set_target_bitmap)
+   {
       new_display->vt->set_target_bitmap(new_display, bitmap);
 
+      /* Set the new shader if necessary.  This should done before the
+       * update_transformation call, which will also update the shader's
+       * al_projview_matrix variable.
+       */
+      if (!same_shader || !new_shader) {
+         al_use_shader(new_shader);
+      }
+
       new_display->vt->update_transformation(new_display, bitmap);
    }
 }
@@ -422,6 +516,19 @@ void al_set_blender(int op, int src, int dst)
 
 
 
+/* Function: al_set_blend_color
+*/
+void al_set_blend_color(ALLEGRO_COLOR color)
+{
+   thread_local_state *tls;
+
+   if ((tls = tls_get()) == NULL)
+      return;
+
+   tls->current_blender.blend_color = color;
+}
+
+
 /* Function: al_set_separate_blender
  */
 void al_set_separate_blender(int op, int src, int dst,
@@ -461,6 +568,19 @@ void al_get_blender(int *op, int *src, int *dst)
 
 
 
+/* Function: al_get_blend_color
+*/
+ALLEGRO_COLOR al_get_blend_color(void)
+{
+   thread_local_state *tls;
+
+   if ((tls = tls_get()) == NULL)
+      return al_map_rgba(255, 255, 255, 255);
+
+   return tls->current_blender.blend_color;
+}
+
+
 /* Function: al_get_separate_blender
  */
 void al_get_separate_blender(int *op, int *src, int *dst,
@@ -517,10 +637,6 @@ void al_set_new_bitmap_flags(int flags)
    if ((tls = tls_get()) == NULL)
       return;
 
-   /* Assume ALLEGRO_VIDEO_BITMAP if ALLEGRO_MEMORY_BITMAP is not set. */
-   if (!(flags & ALLEGRO_MEMORY_BITMAP))
-      flags |= ALLEGRO_VIDEO_BITMAP;
-
    tls->new_bitmap_flags = flags;
 }
 
@@ -565,7 +681,6 @@ int al_get_new_bitmap_flags(void)
 
 
 
-#define _STORE(x) stored->tls.x = tls->x;
 /* Function: al_store_state
  */
 void al_store_state(ALLEGRO_STATE *state, int flags)
@@ -579,6 +694,8 @@ void al_store_state(ALLEGRO_STATE *state, int flags)
    stored = (void *)state;
    stored->flags = flags;
 
+#define _STORE(x) (stored->tls.x = tls->x)
+
    if (flags & ALLEGRO_STATE_NEW_DISPLAY_PARAMETERS) {
       _STORE(new_display_flags);
       _STORE(new_display_refresh_rate);
@@ -586,13 +703,15 @@ void al_store_state(ALLEGRO_STATE *state, int flags)
       _STORE(new_window_x);
       _STORE(new_window_y);
       _STORE(new_display_settings);
+      _al_sane_strncpy(stored->tls.new_window_title, tls->new_window_title,
+                       strlen(tls->new_window_title));
    }
-   
+
    if (flags & ALLEGRO_STATE_NEW_BITMAP_PARAMETERS) {
       _STORE(new_bitmap_format);
       _STORE(new_bitmap_flags);
    }
-   
+
    if (flags & ALLEGRO_STATE_DISPLAY) {
       _STORE(current_display);
    }
@@ -609,7 +728,7 @@ void al_store_state(ALLEGRO_STATE *state, int flags)
       _STORE(new_file_interface);
       _STORE(fs_interface);
    }
-   
+
    if (flags & ALLEGRO_STATE_TRANSFORM) {
       ALLEGRO_BITMAP *target = al_get_target_bitmap();
       if (!target)
@@ -617,12 +736,19 @@ void al_store_state(ALLEGRO_STATE *state, int flags)
       else
          stored->stored_transform = target->transform;
    }
-}
+
+   if (flags & ALLEGRO_STATE_PROJECTION_TRANSFORM) {
+      ALLEGRO_BITMAP *target = al_get_target_bitmap();
+      if (target) {
+         stored->stored_projection_transform = target->proj_transform;
+      }
+   }
+
 #undef _STORE
+}
 
 
 
-#define _STORE(x) tls->x = stored->tls.x;
 /* Function: al_restore_state
  */
 void al_restore_state(ALLEGRO_STATE const *state)
@@ -633,50 +759,65 @@ void al_restore_state(ALLEGRO_STATE const *state)
 
    if ((tls = tls_get()) == NULL)
       return;
-   
+
    stored = (void *)state;
    flags = stored->flags;
 
+#define _RESTORE(x) (tls->x = stored->tls.x)
+
    if (flags & ALLEGRO_STATE_NEW_DISPLAY_PARAMETERS) {
-      _STORE(new_display_flags);
-      _STORE(new_display_refresh_rate);
-      _STORE(new_display_adapter);
-      _STORE(new_window_x);
-      _STORE(new_window_y);
-      _STORE(new_display_settings);
+      _RESTORE(new_display_flags);
+      _RESTORE(new_display_refresh_rate);
+      _RESTORE(new_display_adapter);
+      _RESTORE(new_window_x);
+      _RESTORE(new_window_y);
+      _RESTORE(new_display_settings);
+      _al_sane_strncpy(tls->new_window_title, stored->tls.new_window_title,
+                       strlen(tls->new_window_title));
    }
-   
+
    if (flags & ALLEGRO_STATE_NEW_BITMAP_PARAMETERS) {
-      _STORE(new_bitmap_format);
-      _STORE(new_bitmap_flags);
+      _RESTORE(new_bitmap_format);
+      _RESTORE(new_bitmap_flags);
    }
-   
+
    if (flags & ALLEGRO_STATE_DISPLAY) {
-      _STORE(current_display);
-      _al_set_current_display_only(tls->current_display);
+      if (tls->current_display != stored->tls.current_display) {
+         _al_set_current_display_only(stored->tls.current_display);
+         ASSERT(tls->current_display == stored->tls.current_display);
+      }
    }
 
    if (flags & ALLEGRO_STATE_TARGET_BITMAP) {
-      _STORE(target_bitmap);
-      al_set_target_bitmap(tls->target_bitmap);
+      if (tls->target_bitmap != stored->tls.target_bitmap) {
+         al_set_target_bitmap(stored->tls.target_bitmap);
+         ASSERT(tls->target_bitmap == stored->tls.target_bitmap);
+      }
    }
-   
+
    if (flags & ALLEGRO_STATE_BLENDER) {
       tls->current_blender = stored->stored_blender;
    }
 
    if (flags & ALLEGRO_STATE_NEW_FILE_INTERFACE) {
-      _STORE(new_file_interface);
-      _STORE(fs_interface);
+      _RESTORE(new_file_interface);
+      _RESTORE(fs_interface);
    }
-   
+
    if (flags & ALLEGRO_STATE_TRANSFORM) {
       ALLEGRO_BITMAP *bitmap = al_get_target_bitmap();
       if (bitmap)
          al_use_transform(&stored->stored_transform);
    }
+
+   if (flags & ALLEGRO_STATE_PROJECTION_TRANSFORM) {
+      ALLEGRO_BITMAP *bitmap = al_get_target_bitmap();
+      if (bitmap)
+         al_use_projection_transform(&stored->stored_projection_transform);
+   }
+
+#undef _RESTORE
 }
-#undef _STORE
 
 
 
@@ -782,6 +923,28 @@ void al_set_errno(int errnum)
 }
 
 
+#ifdef ALLEGRO_ANDROID
+JNIEnv *_al_android_get_jnienv(void)
+{
+   thread_local_state *tls;
+
+   if ((tls = tls_get()) == NULL)
+      return 0;
+   return tls->jnienv;
+}
+
+
+
+void _al_android_set_jnienv(JNIEnv *jnienv)
+{
+   thread_local_state *tls;
+
+   if ((tls = tls_get()) == NULL)
+      return;
+   tls->jnienv = jnienv;
+}
+#endif
+
 
 int *_al_tls_get_dtor_owner_count(void)
 {
@@ -792,5 +955,4 @@ int *_al_tls_get_dtor_owner_count(void)
 }
 
 
-
 /* vim: set sts=3 sw=3 et: */
