@@ -1,6 +1,6 @@
-/*         ______   ___    ___ 
- *        /\  _  \ /\_ \  /\_ \ 
- *        \ \ \L\ \\//\ \ \//\ \      __     __   _ __   ___ 
+/*         ______   ___    ___
+ *        /\  _  \ /\_ \  /\_ \
+ *        \ \ \L\ \\//\ \ \//\ \      __     __   _ __   ___
  *         \ \  __ \ \ \ \  \ \ \   /'__`\ /'_ `\/\`'__\/ __`\
  *          \ \ \/\ \ \_\ \_ \_\ \_/\  __//\ \L\ \ \ \//\ \L\ \
  *           \ \_\ \_\/\____\/\____\ \____\ \____ \ \_\\ \____/
@@ -20,6 +20,8 @@
 #define WINVER 0x0501
 #endif
 
+#define UNICODE
+
 #include <windows.h>
 #include <windowsx.h>
 
@@ -36,6 +38,7 @@
 #include "allegro5/internal/aintern_bitmap.h"
 #include "allegro5/internal/aintern_vector.h"
 #include "allegro5/internal/aintern_display.h"
+#include "allegro5/internal/aintern_wunicode.h"
 #include "allegro5/platform/aintwin.h"
 
 
@@ -51,6 +54,10 @@ UINT _al_win_msg_call_proc = 0;
 UINT _al_win_msg_suicide = 0;
 
 
+#ifndef WM_DPICHANGED
+#define WM_DPICHANGED 0x02E0
+#endif
+
 
 static void display_flags_to_window_styles(int flags,
    DWORD *style, DWORD *ex_style)
@@ -71,8 +78,8 @@ static void display_flags_to_window_styles(int flags,
 
 HWND _al_win_create_hidden_window()
 {
-   HWND window = CreateWindowEx(0, 
-      "ALEX", "hidden", WS_POPUP,
+   HWND window = CreateWindowEx(0,
+      L"ALEX", L"hidden", WS_POPUP,
       -5000, -5000, 0, 0,
       NULL,NULL,window_class.hInstance,0);
    return window;
@@ -130,6 +137,7 @@ HWND _al_win_create_window(ALLEGRO_DISPLAY *display, int width, int height, int
    ALLEGRO_DISPLAY_WIN *win_display = (ALLEGRO_DISPLAY_WIN *)display;
    WINDOWINFO wi;
    int lsize, rsize, tsize, bsize; // left, right, top, bottom border sizes
+   wchar_t* window_title;
 
    wi.cbSize = sizeof(WINDOWINFO);
 
@@ -148,10 +156,15 @@ HWND _al_win_create_window(ALLEGRO_DISPLAY *display, int width, int height, int
       _al_win_get_window_center(win_display, width, height, &pos_x, &pos_y);
    }
 
+   window_title = _al_win_utf16(al_get_new_window_title());
    my_window = CreateWindowEx(ex_style,
-      "ALEX", "Allegro", style,
+      L"ALEX", window_title, style,
       pos_x, pos_y, width, height,
       NULL,NULL,window_class.hInstance,0);
+   al_free(window_title);
+
+   if (_al_win_register_touch_window)
+      _al_win_register_touch_window(my_window, 0);
 
    GetWindowInfo(my_window, &wi);
 
@@ -185,29 +198,39 @@ HWND _al_win_create_window(ALLEGRO_DISPLAY *display, int width, int height, int
       DrawMenuBar(my_window);
    }
 
+   _al_vector_init(&win_display->msg_callbacks, sizeof(ALLEGRO_DISPLAY_WIN_CALLBACK));
+
    return my_window;
 }
 
 
-HWND _al_win_create_faux_fullscreen_window(LPCTSTR devname, ALLEGRO_DISPLAY *display, 
-	int x1, int y1, int width, int height, int refresh_rate, int flags)
-{ 
+HWND _al_win_create_faux_fullscreen_window(LPCTSTR devname, ALLEGRO_DISPLAY *display,
+    int x1, int y1, int width, int height, int refresh_rate, int flags)
+{
    HWND my_window;
    DWORD style;
    DWORD ex_style;
    DEVMODE mode;
    LONG temp;
+   wchar_t* window_title;
 
-   (void)display;
+   ALLEGRO_DISPLAY_WIN *win_display = (ALLEGRO_DISPLAY_WIN *)display;
    (void)flags;
 
+   _al_vector_init(&win_display->msg_callbacks, sizeof(ALLEGRO_DISPLAY_WIN_CALLBACK));
+
    style = WS_VISIBLE;
    ex_style = WS_EX_TOPMOST;
 
+   window_title = _al_win_utf16(al_get_new_window_title());
    my_window = CreateWindowEx(ex_style,
-      "ALEX", "Allegro", style,
+      L"ALEX", window_title, style,
       x1, y1, width, height,
       NULL,NULL,window_class.hInstance,0);
+   al_free(window_title);
+
+   if (_al_win_register_touch_window)
+      _al_win_register_touch_window(my_window, 0);
 
    temp = GetWindowLong(my_window, GWL_STYLE);
    temp &= ~WS_CAPTION;
@@ -219,17 +242,17 @@ HWND _al_win_create_faux_fullscreen_window(LPCTSTR devname, ALLEGRO_DISPLAY *dis
    mode.dmSize = sizeof(DEVMODE);
    mode.dmDriverExtra = 0;
    mode.dmBitsPerPel = al_get_new_display_option(ALLEGRO_COLOR_SIZE, NULL);
-   mode.dmPelsWidth = width; 
+   mode.dmPelsWidth = width;
    mode.dmPelsHeight = height;
    mode.dmDisplayFlags = 0;
    mode.dmDisplayFrequency = refresh_rate;
    mode.dmPosition.x = x1;
    mode.dmPosition.y = y1;
    mode.dmFields = DM_BITSPERPEL|DM_PELSWIDTH|DM_PELSHEIGHT|DM_DISPLAYFLAGS|
-   	DM_DISPLAYFREQUENCY|DM_POSITION;
+    DM_DISPLAYFREQUENCY|DM_POSITION;
 
    ChangeDisplaySettingsEx(devname, &mode, NULL, 0, NULL/*CDS_FULLSCREEN*/);
-   
+
    return my_window;
 }
 
@@ -346,27 +369,44 @@ static void break_window_message_pump(ALLEGRO_DISPLAY_WIN *win_display, HWND hWn
    PostThreadMessage(wnd_thread_id, WM_NULL, 0, 0);
 }
 
-static LRESULT CALLBACK window_callback(HWND hWnd, UINT message, 
+/* Windows Touch Input emulate WM_MOUSE* events. If we are using touch input explicitly,
+ * we do not want to use this, because Allegro driver provide emulation already. This
+ * way we can be consistent across platforms.
+ */
+static bool accept_mouse_event(void)
+{
+   if (!al_is_touch_input_installed())
+      return true;
+
+   return !((GetMessageExtraInfo() & _AL_MOUSEEVENTF_FROMTOUCH) == _AL_MOUSEEVENTF_FROMTOUCH);
+}
+
+static LRESULT CALLBACK window_callback(HWND hWnd, UINT message,
     WPARAM wParam, LPARAM lParam)
 {
    ALLEGRO_DISPLAY *d = NULL;
    ALLEGRO_DISPLAY_WIN *win_display = NULL;
+   //WINDOWINFO wi;
    unsigned int i;
    ALLEGRO_EVENT_SOURCE *es = NULL;
    ALLEGRO_SYSTEM *system = al_get_system_driver();
 
+   //wi.cbSize = sizeof(WINDOWINFO);
+
    if (message == _al_win_msg_call_proc) {
       ((void (*)(void*))wParam) ((void*)lParam);
       return 0;
    }
 
    if (!system) {
-      return DefWindowProc(hWnd,message,wParam,lParam); 
+      return DefWindowProc(hWnd,message,wParam,lParam);
    }
 
    if (message == _al_win_msg_suicide && wParam) {
       win_display = (ALLEGRO_DISPLAY_WIN*)wParam;
       break_window_message_pump(win_display, hWnd);
+      if (_al_win_unregister_touch_window)
+         _al_win_unregister_touch_window(hWnd);
       DestroyWindow(hWnd);
       return 0;
    }
@@ -382,16 +422,25 @@ static LRESULT CALLBACK window_callback(HWND hWnd, UINT message,
    }
 
    if (i == system->displays._size)
-      return DefWindowProc(hWnd,message,wParam,lParam); 
+      return DefWindowProc(hWnd,message,wParam,lParam);
 
    if (message == _al_win_msg_suicide) {
       break_window_message_pump(win_display, hWnd);
+      if (_al_win_unregister_touch_window)
+         _al_win_unregister_touch_window(hWnd);
       DestroyWindow(hWnd);
       return 0;
    }
 
+   for (i = 0; i < _al_vector_size(&win_display->msg_callbacks); ++i) {
+      LRESULT result = TRUE;
+      ALLEGRO_DISPLAY_WIN_CALLBACK *ptr = _al_vector_ref(&win_display->msg_callbacks, i);
+      if ((ptr->proc)(d, message, wParam, lParam, &result, ptr->userdata))
+         return result;
+   }
+
    switch (message) {
-      case WM_INPUT: 
+      case WM_INPUT:
       {
          /* RAW Input is currently unused. */
           UINT dwSize;
@@ -402,17 +451,17 @@ static LRESULT CALLBACK window_callback(HWND hWnd, UINT message,
           if (!al_is_mouse_installed())
              break;
 
-          GetRawInputData((HRAWINPUT)lParam, RID_INPUT, NULL, &dwSize, 
+          GetRawInputData((HRAWINPUT)lParam, RID_INPUT, NULL, &dwSize,
                           sizeof(RAWINPUTHEADER));
           lpb = al_malloc(sizeof(BYTE)*dwSize);
-          if (lpb == NULL) 
+          if (lpb == NULL)
               break;
 
           GetRawInputData((HRAWINPUT)lParam, RID_INPUT, lpb, &dwSize, sizeof(RAWINPUTHEADER));
           raw = (RAWINPUT*)lpb;
 
           if (raw->header.dwType != RIM_TYPEMOUSE) {
-             al_free(lpb); 
+             al_free(lpb);
              break;
           }
 
@@ -448,83 +497,100 @@ static LRESULT CALLBACK window_callback(HWND hWnd, UINT message,
 
           if (rm->usButtonFlags & RI_MOUSE_WHEEL) {
              SHORT z = (SHORT)rm->usButtonData;
-             _al_win_mouse_handle_wheel(z / WHEEL_DELTA, false, win_display);
+             _al_win_mouse_handle_wheel(z, false, win_display);
           }
        }
 
-          al_free(lpb); 
+          al_free(lpb);
           break;
       }
       case WM_LBUTTONDOWN:
       case WM_LBUTTONUP: {
-         int mx = GET_X_LPARAM(lParam);
-         int my = GET_Y_LPARAM(lParam);
-         bool down = (message == WM_LBUTTONDOWN);
-         _al_win_mouse_handle_button(1, down, mx, my, true, win_display);
-         handle_mouse_capture(down, hWnd);
+         if (accept_mouse_event()) {
+            int mx = GET_X_LPARAM(lParam);
+            int my = GET_Y_LPARAM(lParam);
+            bool down = (message == WM_LBUTTONDOWN);
+            _al_win_mouse_handle_button(1, down, mx, my, true, win_display);
+            handle_mouse_capture(down, hWnd);
+         }
          break;
       }
       case WM_MBUTTONDOWN:
       case WM_MBUTTONUP: {
-         int mx = GET_X_LPARAM(lParam);
-         int my = GET_Y_LPARAM(lParam);
-         bool down = (message == WM_MBUTTONDOWN);
-         _al_win_mouse_handle_button(3, down, mx, my, true, win_display);
-         handle_mouse_capture(down, hWnd);
+         if (accept_mouse_event()) {
+            int mx = GET_X_LPARAM(lParam);
+            int my = GET_Y_LPARAM(lParam);
+            bool down = (message == WM_MBUTTONDOWN);
+            _al_win_mouse_handle_button(3, down, mx, my, true, win_display);
+            handle_mouse_capture(down, hWnd);
+         }
          break;
       }
       case WM_RBUTTONDOWN:
       case WM_RBUTTONUP: {
-         int mx = GET_X_LPARAM(lParam);
-         int my = GET_Y_LPARAM(lParam);
-         bool down = (message == WM_RBUTTONDOWN);
-         _al_win_mouse_handle_button(2, down, mx, my, true, win_display);
-         handle_mouse_capture(down, hWnd);
+         if (accept_mouse_event()) {
+            int mx = GET_X_LPARAM(lParam);
+            int my = GET_Y_LPARAM(lParam);
+            bool down = (message == WM_RBUTTONDOWN);
+            _al_win_mouse_handle_button(2, down, mx, my, true, win_display);
+            handle_mouse_capture(down, hWnd);
+         }
          break;
       }
       case WM_XBUTTONDOWN:
       case WM_XBUTTONUP: {
-         int mx = GET_X_LPARAM(lParam);
-         int my = GET_Y_LPARAM(lParam);
-         int button = HIWORD(wParam);
-         bool down = (message == WM_XBUTTONDOWN);
-         if (button == XBUTTON1)
-            _al_win_mouse_handle_button(4, down, mx, my, true, win_display);
-         else if (button == XBUTTON2)
-            _al_win_mouse_handle_button(5, down, mx, my, true, win_display);
-         handle_mouse_capture(down, hWnd);
-         return TRUE;
+         if (accept_mouse_event()) {
+            int mx = GET_X_LPARAM(lParam);
+            int my = GET_Y_LPARAM(lParam);
+            int button = HIWORD(wParam);
+            bool down = (message == WM_XBUTTONDOWN);
+            if (button == XBUTTON1)
+               _al_win_mouse_handle_button(4, down, mx, my, true, win_display);
+            else if (button == XBUTTON2)
+               _al_win_mouse_handle_button(5, down, mx, my, true, win_display);
+            handle_mouse_capture(down, hWnd);
+            return TRUE;
+         }
+         break;
       }
       case WM_MOUSEWHEEL: {
-         int d = GET_WHEEL_DELTA_WPARAM(wParam);
-         _al_win_mouse_handle_wheel(d / WHEEL_DELTA, false, win_display);
-         return TRUE;
+         if (accept_mouse_event()) {
+            int d = GET_WHEEL_DELTA_WPARAM(wParam);
+            _al_win_mouse_handle_wheel(d, false, win_display);
+            return TRUE;
+         }
+         break;
       }
       case WM_MOUSEHWHEEL: {
-         int d = GET_WHEEL_DELTA_WPARAM(wParam);
-         _al_win_mouse_handle_hwheel(d / WHEEL_DELTA, false, win_display);
-         return TRUE;
+         if (accept_mouse_event()) {
+            int d = GET_WHEEL_DELTA_WPARAM(wParam);
+            _al_win_mouse_handle_hwheel(d, false, win_display);
+            return TRUE;
+         }
+         break;
       }
       case WM_MOUSEMOVE: {
-         TRACKMOUSEEVENT tme;
-         int mx = GET_X_LPARAM(lParam);
-         int my = GET_Y_LPARAM(lParam);
+         if (accept_mouse_event()) {
+            TRACKMOUSEEVENT tme;
+            int mx = GET_X_LPARAM(lParam);
+            int my = GET_Y_LPARAM(lParam);
 
-         if (win_display->mouse_cursor_shown && we_hid_the_mouse) {
-            we_hid_the_mouse = false;
-            win_display->display.vt->hide_mouse_cursor((void*)win_display);
-         }
+            if (win_display->mouse_cursor_shown && we_hid_the_mouse) {
+               we_hid_the_mouse = false;
+               win_display->display.vt->hide_mouse_cursor((void*)win_display);
+            }
 
-         _al_win_mouse_handle_move(mx, my, true, win_display);
-         if (mx >= 0 && my >= 0 && mx < d->w && my < d->h) {
-            tme.cbSize = sizeof(tme);
-            tme.dwFlags = TME_QUERY;
-            if (TrackMouseEvent(&tme) && !tme.hwndTrack) {
-               tme.dwFlags = TME_LEAVE;
-               tme.hwndTrack = hWnd;
-               tme.dwHoverTime = 0;
-               TrackMouseEvent(&tme);
-               _al_win_mouse_handle_enter(win_display);
+            _al_win_mouse_handle_move(mx, my, true, win_display);
+            if (mx >= 0 && my >= 0 && mx < d->w && my < d->h) {
+               tme.cbSize = sizeof(tme);
+               tme.dwFlags = TME_QUERY;
+               if (TrackMouseEvent(&tme) && !tme.hwndTrack) {
+                  tme.dwFlags = TME_LEAVE;
+                  tme.hwndTrack = hWnd;
+                  tme.dwHoverTime = 0;
+                  TrackMouseEvent(&tme);
+                  _al_win_mouse_handle_enter(win_display);
+               }
             }
          }
 
@@ -540,7 +606,52 @@ static LRESULT CALLBACK window_callback(HWND hWnd, UINT message,
          break;
       }
       case WM_MOUSELEAVE: {
-         _al_win_mouse_handle_leave(win_display);
+         if (accept_mouse_event()) {
+            _al_win_mouse_handle_leave(win_display);
+         }
+         break;
+      }
+      case _AL_WM_TOUCH: {
+         if (_al_win_get_touch_input_info && _al_win_close_touch_input_handle) {
+            int number_of_touches = LOWORD(wParam);
+
+            TOUCHINPUT* touches = al_malloc(number_of_touches * sizeof(TOUCHINPUT));
+
+            if (_al_win_get_touch_input_info((HANDLE)lParam, number_of_touches, touches, sizeof(TOUCHINPUT))) {
+
+               if (al_is_touch_input_installed()) {
+
+                  int i;
+
+                  POINT origin = { 0, 0 };
+
+                  ClientToScreen(hWnd, &origin);
+
+                  _al_win_touch_input_set_time_stamp((touches + number_of_touches - 1)->dwTime);
+
+                  for (i = 0; i < number_of_touches; ++i) {
+
+                     TOUCHINPUT* touch = touches + i;
+
+                     float x = touch->x / 100.0f - (float)origin.x;
+                     float y = touch->y / 100.0f - (float)origin.y;
+
+                     bool primary = touch->dwFlags & _AL_TOUCHEVENTF_PRIMARY ? true : false;
+
+                     if (touch->dwFlags & _AL_TOUCHEVENTF_DOWN)
+                        _al_win_touch_input_handle_begin((int)touch->dwID, (size_t)touch->dwTime, x, y, primary, win_display);
+                     else if (touch->dwFlags & _AL_TOUCHEVENTF_UP)
+                        _al_win_touch_input_handle_end((int)touch->dwID, (size_t)touch->dwTime, x, y, primary, win_display);
+                     else if (touch->dwFlags & _AL_TOUCHEVENTF_MOVE)
+                        _al_win_touch_input_handle_move((int)touch->dwID, (size_t)touch->dwTime, x, y, primary, win_display);
+                  }
+               }
+
+               _al_win_close_touch_input_handle((HANDLE)lParam);
+            }
+
+            al_free(touches);
+         }
          break;
       }
       case WM_CAPTURECHANGED: {
@@ -572,7 +683,7 @@ static LRESULT CALLBACK window_callback(HWND hWnd, UINT message,
          break;
       }
       case WM_KEYDOWN: {
-         int vcode = wParam; 
+         int vcode = wParam;
          int scode = (lParam >> 16) & 0xff;
          bool extended = (lParam >> 24) & 0x1;
          bool repeated = (lParam >> 30) & 0x1;
@@ -620,6 +731,7 @@ static LRESULT CALLBACK window_callback(HWND hWnd, UINT message,
                GetRegionData(hrgn, size, rgndata);
                n = rgndata->rdh.nCount;
                rects = (RECT *)rgndata->Buffer;
+               //GetWindowInfo(win_display->window, &wi);
                _al_event_source_lock(es);
                if (_al_event_source_needs_to_generate_event(es)) {
                   ALLEGRO_EVENT event;
@@ -706,8 +818,6 @@ static LRESULT CALLBACK window_callback(HWND hWnd, UINT message,
                SetWindowPos(win_display->window, HWND_NOTOPMOST, 0, 0, 0, 0, SWP_NOMOVE | SWP_NOSIZE);
                SetWindowPos(GetForegroundWindow(), HWND_TOP, 0, 0, 0, 0, SWP_NOMOVE | SWP_NOSIZE);
             }
-            // Show the taskbar in case we hid it
-            SetWindowPos(FindWindow("Shell_traywnd", ""), 0, 0, 0, 0, 0, SWP_SHOWWINDOW);
             if (d->flags & ALLEGRO_FULLSCREEN) {
                d->vt->switch_out(d);
             }
@@ -736,6 +846,39 @@ static LRESULT CALLBACK window_callback(HWND hWnd, UINT message,
          }
          _al_event_source_unlock(es);
          return 0;
+      case WM_GETMINMAXINFO:
+         {
+            LPMINMAXINFO p_info = (LPMINMAXINFO)lParam;
+            RECT wRect;
+            RECT cRect;
+            int wWidth;
+            int wHeight;
+            int cWidth;
+            int cHeight;
+
+            GetWindowRect(hWnd, &wRect);
+            GetClientRect(hWnd, &cRect);
+            wWidth = wRect.right - wRect.left;
+            wHeight = wRect.bottom - wRect.top;
+            cWidth = cRect.right - cRect.left;
+            cHeight = cRect.bottom - cRect.top;
+
+            /* Client size is zero when the window is restored. */
+            if (cWidth != 0 && cHeight != 0) {
+               int total_border_width = wWidth - cWidth;
+               int total_border_height = wHeight - cHeight;
+               POINT wmin, wmax;
+
+               wmin.x = (d->min_w > 0) ? d->min_w + total_border_width : p_info->ptMinTrackSize.x;
+               wmin.y = (d->min_h > 0) ? d->min_h + total_border_height : p_info->ptMinTrackSize.y;
+               wmax.x = (d->max_w > 0) ? d->max_w + total_border_width : p_info->ptMaxTrackSize.x;
+               wmax.y = (d->max_h > 0) ? d->max_h + total_border_height : p_info->ptMaxTrackSize.y;
+
+               p_info->ptMinTrackSize = wmin;
+               p_info->ptMaxTrackSize = wmax;
+            }
+         }
+         break;
       case WM_SIZE:
          if (wParam == SIZE_RESTORED || wParam == SIZE_MAXIMIZED || wParam == SIZE_MINIMIZED) {
             /*
@@ -745,6 +888,10 @@ static LRESULT CALLBACK window_callback(HWND hWnd, UINT message,
                resize_postponed = true;
                _beginthread(postpone_thread_proc, 0, (void *)d);
             }
+            d->flags &= ~ALLEGRO_MAXIMIZED;
+            if (wParam == SIZE_MAXIMIZED) {
+                d->flags |= ALLEGRO_MAXIMIZED;
+            }
          }
          return 0;
       case WM_ENTERSIZEMOVE:
@@ -767,30 +914,45 @@ static LRESULT CALLBACK window_callback(HWND hWnd, UINT message,
             win_display->can_acknowledge = true;
          }
          break;
-   } 
+      case WM_DPICHANGED:
+        if ((d->flags & ALLEGRO_RESIZABLE) && !(d->flags & ALLEGRO_FULLSCREEN)) {
+            RECT* rect = (RECT*)lParam;
+            // XXX: This doesn't seem to actually move the window... why?
+            SetWindowPos(
+               hWnd,
+               0,
+               rect->left,
+               rect->top,
+               rect->right - rect->left,
+               rect->bottom - rect->top,
+               SWP_NOZORDER | SWP_NOACTIVATE);
+            win_generate_resize_event(win_display);
+        }
+        break;
+   }
 
-   return DefWindowProc(hWnd,message,wParam,lParam); 
+   return DefWindowProc(hWnd,message,wParam,lParam);
 }
 
 int _al_win_init_window()
 {
-   // Create A Window Class Structure 
+   // Create A Window Class Structure
    window_class.cbClsExtra = 0;
-   window_class.cbWndExtra = 0; 
+   window_class.cbWndExtra = 0;
    window_class.hbrBackground = NULL;
-   window_class.hCursor = NULL; 
-   window_class.hIcon = NULL; 
+   window_class.hCursor = NULL;
+   window_class.hIcon = NULL;
    window_class.hInstance = GetModuleHandle(NULL);
    window_class.lpfnWndProc = window_callback;
-   window_class.lpszClassName = "ALEX";
+   window_class.lpszClassName = L"ALEX";
    window_class.lpszMenuName = NULL;
    window_class.style = CS_VREDRAW|CS_HREDRAW|CS_OWNDC;
 
    RegisterClass(&window_class);
 
    if (_al_win_msg_call_proc == 0 && _al_win_msg_suicide == 0) {
-      _al_win_msg_call_proc = RegisterWindowMessage("Allegro call proc");
-      _al_win_msg_suicide = RegisterWindowMessage("Allegro window suicide");
+      _al_win_msg_call_proc = RegisterWindowMessage(L"Allegro call proc");
+      _al_win_msg_suicide = RegisterWindowMessage(L"Allegro window suicide");
    }
 
    return true;
@@ -899,6 +1061,18 @@ void _al_win_set_display_icons(ALLEGRO_DISPLAY *display,
       num_icons, bmps);
 }
 
+void _al_win_destroy_display_icons(ALLEGRO_DISPLAY *display)
+{
+   ALLEGRO_DISPLAY_WIN *win_display = (ALLEGRO_DISPLAY_WIN *)display;
+   HICON old_icon;
+
+   old_icon = (HICON)SendMessage(win_display->window, WM_SETICON, ICON_SMALL, 0);
+   if (old_icon)
+      DestroyIcon(old_icon);
+   old_icon = (HICON)SendMessage(win_display->window, WM_SETICON, ICON_BIG, 0);
+   if (old_icon)
+      DestroyIcon(old_icon);
+}
 
 void _al_win_set_window_position(HWND window, int x, int y)
 {
@@ -1022,16 +1196,6 @@ bool _al_win_set_display_flag(ALLEGRO_DISPLAY *display, int flag, bool onoff)
          if (onoff) {
             // Re-set the TOPMOST flag and move to position
             SetWindowPos(win_display->window, HWND_TOPMOST, mi.x1, mi.y1, 0, 0, SWP_NOSIZE);
-
-            // Hide the taskbar if fullscreening on primary monitor
-            if (win_display->adapter == 0) {
-               SetWindowPos(
-                  FindWindow("Shell_traywnd", ""),
-                  0,
-                  0, 0,
-                  0, 0,
-                  SWP_HIDEWINDOW | SWP_NOSIZE | SWP_NOZORDER | SWP_NOMOVE);
-            }
          }
          else {
             int pos_x = 0;
@@ -1042,9 +1206,6 @@ bool _al_win_set_display_flag(ALLEGRO_DISPLAY *display, int flag, bool onoff)
             // Unset the topmost flag
             SetWindowPos(win_display->window, HWND_NOTOPMOST, 0, 0, 0, 0, SWP_NOMOVE | SWP_NOSIZE);
 
-            // Show the taskbar
-            SetWindowPos(
-               FindWindow("Shell_traywnd", ""), 0, 0, 0, 0, 0, SWP_SHOWWINDOW | SWP_NOSIZE | SWP_NOZORDER | SWP_NOMOVE);
             // Center the window
             _al_win_get_window_center(win_display, display->w, display->h, &pos_x, &pos_y);
             GetWindowInfo(win_display->window, &wi);
@@ -1062,6 +1223,16 @@ bool _al_win_set_display_flag(ALLEGRO_DISPLAY *display, int flag, bool onoff)
 
          ASSERT(!!(display->flags & ALLEGRO_FULLSCREEN_WINDOW) == onoff);
          return true;
+      case ALLEGRO_MAXIMIZED:
+         if ((!!(display->flags & ALLEGRO_MAXIMIZED)) == onoff)
+            return true;
+         if (onoff) {
+            ShowWindow(win_display->window, SW_SHOWMAXIMIZED);
+         }
+         else {
+            ShowWindow(win_display->window, SW_RESTORE);
+         }
+         return true;
    }
    return false;
 }
@@ -1070,7 +1241,48 @@ bool _al_win_set_display_flag(ALLEGRO_DISPLAY *display, int flag, bool onoff)
 void _al_win_set_window_title(ALLEGRO_DISPLAY *display, const char *title)
 {
    ALLEGRO_DISPLAY_WIN *win_display = (ALLEGRO_DISPLAY_WIN *)display;
-   SetWindowText(win_display->window, title);
+   wchar_t *utf16 = _al_win_utf16(title);
+   SetWindowText(win_display->window, utf16);
+   al_free(utf16);
+}
+
+bool _al_win_set_window_constraints(ALLEGRO_DISPLAY *display,
+   int min_w, int min_h, int max_w, int max_h)
+{
+   ALLEGRO_DISPLAY_WIN *win_display = (ALLEGRO_DISPLAY_WIN *)display;
+   win_display->display.min_w = min_w;
+   win_display->display.min_h = min_h;
+   win_display->display.max_w = max_w;
+   win_display->display.max_h = max_h;
+
+   /* Resize so constraints can take effect. */
+   al_resize_display(display, display->w, display->h);
+
+   return true;
+}
+
+void _al_win_post_create_window(ALLEGRO_DISPLAY *display)
+{
+   /* Ideally the d3d/wgl window creation would already create the
+    * window maximized - but that code looks too messy to me to touch
+    * right now.
+    */
+   if (display->flags & ALLEGRO_MAXIMIZED) {
+      display->flags &= ~ALLEGRO_MAXIMIZED;
+      al_set_display_flag(display, ALLEGRO_MAXIMIZED, true);
+   }
+}
+
+bool _al_win_get_window_constraints(ALLEGRO_DISPLAY *display,
+   int *min_w, int *min_h, int *max_w, int *max_h)
+{
+   ALLEGRO_DISPLAY_WIN *win_display = (ALLEGRO_DISPLAY_WIN *)display;
+   *min_w = win_display->display.min_w;
+   *min_h = win_display->display.min_h;
+   *max_w = win_display->display.max_w;
+   *max_h = win_display->display.max_h;
+
+   return true;
 }
 
 
@@ -1126,4 +1338,55 @@ int _al_win_determine_adapter(void)
    return a;
 }
 
+/* Function: al_win_add_window_callback
+ */
+bool al_win_add_window_callback(ALLEGRO_DISPLAY *display,
+   bool (*callback)(ALLEGRO_DISPLAY *display, UINT message, WPARAM wparam,
+   LPARAM lparam, LRESULT *result, void *userdata), void *userdata)
+{
+   ALLEGRO_DISPLAY_WIN *win_display = (ALLEGRO_DISPLAY_WIN *) display;
+   ALLEGRO_DISPLAY_WIN_CALLBACK *ptr;
+   
+   if (!display || !callback) {
+      return false;
+   }
+   else {
+      size_t i;
+      for (i = 0; i < _al_vector_size(&win_display->msg_callbacks); ++i) {
+         ALLEGRO_DISPLAY_WIN_CALLBACK *ptr = _al_vector_ref(&win_display->msg_callbacks, i);
+         if (ptr->proc == callback && ptr->userdata == userdata)
+            return false;
+      }
+   }
+
+   if (!(ptr = _al_vector_alloc_back(&win_display->msg_callbacks)))
+      return false;
+
+   ptr->proc = callback;
+   ptr->userdata = userdata;
+   return true;
+}
+
+/* Function: al_win_remove_window_callback
+ */
+bool al_win_remove_window_callback(ALLEGRO_DISPLAY *display,
+   bool (*callback)(ALLEGRO_DISPLAY *display, UINT message, WPARAM wparam,
+   LPARAM lparam, LRESULT *result, void *userdata), void *userdata)
+{
+   ALLEGRO_DISPLAY_WIN *win_display = (ALLEGRO_DISPLAY_WIN *) display;
+   
+   if (display && callback) {
+      size_t i;
+      for (i = 0; i < _al_vector_size(&win_display->msg_callbacks); ++i) {
+         ALLEGRO_DISPLAY_WIN_CALLBACK *ptr = _al_vector_ref(&win_display->msg_callbacks, i);
+         if (ptr->proc == callback && ptr->userdata == userdata) {
+            _al_vector_delete_at(&win_display->msg_callbacks, i);
+            return true;
+         }
+      }
+   }
+
+   return false;
+}
+
 /* vi: set ts=8 sts=3 sw=3 et: */
