@@ -31,6 +31,13 @@
 #include "allegro5/internal/aintern_xmouse.h"
 #include "allegro5/internal/aintern_xsystem.h"
 
+#ifdef ALLEGRO_RASPBERRYPI
+#include "allegro5/internal/aintern_raspberrypi.h"
+#include "allegro5/internal/aintern_vector.h"
+#define ALLEGRO_SYSTEM_XGLX ALLEGRO_SYSTEM_RASPBERRYPI
+#define ALLEGRO_DISPLAY_XGLX ALLEGRO_DISPLAY_RASPBERRYPI
+#endif
+
 ALLEGRO_DEBUG_CHANNEL("mouse")
 
 typedef struct ALLEGRO_MOUSE_XWIN
@@ -63,10 +70,10 @@ static void xmouse_get_state(ALLEGRO_MOUSE_STATE *ret_state);
 static void wheel_motion_handler(int x_button, ALLEGRO_DISPLAY *display);
 static unsigned int x_button_to_al_button(unsigned int x_button);
 static void generate_mouse_event(unsigned int type,
-                                 int x, int y, int z, int w,
-                                 int dx, int dy, int dz, int dw,
-                                 unsigned int button,
-                                 ALLEGRO_DISPLAY *display);
+   int x, int y, int z, int w, float pressure,
+   int dx, int dy, int dz, int dw,
+   unsigned int button,
+   ALLEGRO_DISPLAY *display);
 
 
 
@@ -97,6 +104,27 @@ ALLEGRO_MOUSE_DRIVER *_al_xwin_mouse_driver(void)
 }
 
 
+static void scale_xy(int *x, int *y)
+{
+#ifdef ALLEGRO_RASPBERRYPI
+   ALLEGRO_SYSTEM *s = al_get_system_driver();
+   if (s && s->displays._size > 0) {
+      ALLEGRO_DISPLAY **ref = _al_vector_ref(&s->displays, 0);
+      ALLEGRO_DISPLAY *d = *ref;
+      if (d) {
+         ALLEGRO_DISPLAY_RASPBERRYPI *disp = (void *)d;
+         /* Not sure what's a better approach than adding 0.5 to
+          * get an accurate last pixel */
+         *x = ((*x)+0.5) * d->w / disp->screen_width;
+         *y = ((*y)+0.5) * d->h / disp->screen_height;
+      }
+   }
+#else
+   (void)x;
+   (void)y;
+#endif
+}
+
 
 /* xmouse_init:
  *  Initialise the driver.
@@ -274,7 +302,7 @@ static bool xmouse_set_mouse_axis(int which, int v)
          generate_mouse_event(
             ALLEGRO_EVENT_MOUSE_AXES,
             the_mouse.state.x, the_mouse.state.y, the_mouse.state.z,
-            the_mouse.state.w,
+            the_mouse.state.w, the_mouse.state.pressure,
             0, 0, dz, dw,
             0, the_mouse.state.display);
       }
@@ -325,11 +353,12 @@ void _al_xwin_mouse_button_press_handler(int x_button,
    _al_event_source_lock(&the_mouse.parent.es);
    {
       the_mouse.state.buttons |= (1 << (al_button - 1));
+      the_mouse.state.pressure = the_mouse.state.buttons ? 1.0 : 0.0; /* TODO */
 
       generate_mouse_event(
          ALLEGRO_EVENT_MOUSE_BUTTON_DOWN,
          the_mouse.state.x, the_mouse.state.y, the_mouse.state.z,
-         the_mouse.state.w,
+         the_mouse.state.w, the_mouse.state.pressure,
          0, 0, 0, 0,
          al_button, display);
    }
@@ -351,6 +380,9 @@ static void wheel_motion_handler(int x_button, ALLEGRO_DISPLAY *display)
    if (x_button == 7) dw = 1;
    if (dz == 0 && dw == 0) return;
 
+   dz *= al_get_mouse_wheel_precision();
+   dw *= al_get_mouse_wheel_precision();
+
    _al_event_source_lock(&the_mouse.parent.es);
    {
       the_mouse.state.z += dz;
@@ -359,7 +391,7 @@ static void wheel_motion_handler(int x_button, ALLEGRO_DISPLAY *display)
       generate_mouse_event(
          ALLEGRO_EVENT_MOUSE_AXES,
          the_mouse.state.x, the_mouse.state.y, the_mouse.state.z,
-         the_mouse.state.w,
+         the_mouse.state.w, the_mouse.state.pressure,
          0, 0, dz, dw,
          0, display);
    }
@@ -387,11 +419,12 @@ void _al_xwin_mouse_button_release_handler(int x_button,
    _al_event_source_lock(&the_mouse.parent.es);
    {
       the_mouse.state.buttons &=~ (1 << (al_button - 1));
+      the_mouse.state.pressure = the_mouse.state.buttons ? 1.0 : 0.0; /* TODO */
 
       generate_mouse_event(
          ALLEGRO_EVENT_MOUSE_BUTTON_UP,
          the_mouse.state.x, the_mouse.state.y, the_mouse.state.z,
-         the_mouse.state.w,
+         the_mouse.state.w, the_mouse.state.pressure,
          0, 0, 0, 0,
          al_button, display);
    }
@@ -420,6 +453,8 @@ void _al_xwin_mouse_motion_notify_handler(int x, int y,
    }
 
    _al_event_source_lock(&the_mouse.parent.es);
+   scale_xy(&x, &y);
+
    int dx = x - the_mouse.state.x;
    int dy = y - the_mouse.state.y;
 
@@ -430,7 +465,7 @@ void _al_xwin_mouse_motion_notify_handler(int x, int y,
    generate_mouse_event(
       event_type,
       the_mouse.state.x, the_mouse.state.y, the_mouse.state.z,
-      the_mouse.state.w,
+      the_mouse.state.w, the_mouse.state.pressure,
       dx, dy, 0, 0,
       0, display);
 
@@ -471,7 +506,7 @@ static unsigned int x_button_to_al_button(unsigned int x_button)
  *  Helper to generate a mouse event.
  */
 static void generate_mouse_event(unsigned int type,
-                                 int x, int y, int z, int w,
+                                 int x, int y, int z, int w, float pressure,
                                  int dx, int dy, int dz, int dw,
                                  unsigned int button,
                                  ALLEGRO_DISPLAY *display)
@@ -493,7 +528,7 @@ static void generate_mouse_event(unsigned int type,
    event.mouse.dz = dz;
    event.mouse.dw = dw;
    event.mouse.button = button;
-   event.mouse.pressure = 0.0; /* TODO */
+   event.mouse.pressure = pressure;
    _al_event_source_emit_event(&the_mouse.parent.es, &event);
 }
 
@@ -536,7 +571,7 @@ void _al_xwin_mouse_switch_handler(ALLEGRO_DISPLAY *display,
 
    generate_mouse_event(
       event_type,
-      the_mouse.state.x, the_mouse.state.y, the_mouse.state.z,
+      the_mouse.state.x, the_mouse.state.y, the_mouse.state.z, the_mouse.state.pressure,
       the_mouse.state.w,
       0, 0, 0, 0,
       0, display);
@@ -577,10 +612,5 @@ bool _al_xwin_ungrab_mouse(void)
    return true;
 }
 
-/*
- * Local Variables:
- * c-basic-offset: 3
- * indent-tabs-mode: nil
- * End:
- */
+
 /* vim: set sts=3 sw=3 et: */
