@@ -19,6 +19,7 @@
 #define MAX_TRANS    8
 #define MAX_FONTS    16
 #define MAX_VERTICES 100
+#define MAX_POLYGONS 8
 
 typedef struct {
    ALLEGRO_USTR   *name;
@@ -46,7 +47,7 @@ typedef struct {
 typedef struct {
    ALLEGRO_USTR   *name;
    ALLEGRO_FONT   *font;
-} Font;
+} NamedFont;
 
 int               argc;
 char              **argv;
@@ -55,8 +56,11 @@ ALLEGRO_BITMAP    *membuf;
 Bitmap            bitmaps[MAX_BITMAPS];
 LockRegion        lock_region;
 Transform         transforms[MAX_TRANS];
-Font              fonts[MAX_FONTS];
+NamedFont         fonts[MAX_FONTS];
 ALLEGRO_VERTEX    vertices[MAX_VERTICES];
+float             simple_vertices[2 * MAX_VERTICES];
+int               num_simple_vertices;
+int               vertex_counts[MAX_POLYGONS];
 int               num_global_bitmaps;
 float             delay = 0.0;
 bool              save_outputs = false;
@@ -112,7 +116,7 @@ int               failed_tests = 0;
       (sscanf(stmt, PAT " = " fn " (" PAT##arity " )", lval, ARGS##arity) \
          == 1 + arity)
 
-static void error(char const *msg, ...)
+static void fatal_error(char const *msg, ...)
 {
    va_list ap;
 
@@ -133,19 +137,42 @@ static char const *bmp_type_to_string(BmpType bmp_type)
    return "error";
 }
 
-static ALLEGRO_BITMAP *load_relative_bitmap(char const *filename)
+static ALLEGRO_BITMAP *create_fallback_bitmap(void)
+{
+   ALLEGRO_BITMAP *bmp = al_create_bitmap(256, 256);
+   ALLEGRO_FONT *builtin_font = al_create_builtin_font();
+   ALLEGRO_STATE state;
+
+   if (!bmp) {
+      fatal_error("couldn't create a backup bitmap");
+   }
+   if (!builtin_font) {
+      fatal_error("couldn't create a builtin font");
+   }
+   al_store_state(&state, ALLEGRO_STATE_BITMAP);
+   al_set_target_bitmap(bmp);
+   al_clear_to_color(al_map_rgb_f(0.5, 0, 0));
+   al_draw_text(builtin_font, al_map_rgb_f(1, 1, 1), 0, 0, 0, "fallback");
+   al_restore_state(&state);
+   al_destroy_font(builtin_font);
+
+   return bmp;
+}
+
+static ALLEGRO_BITMAP *load_relative_bitmap(char const *filename, int flags)
 {
    ALLEGRO_BITMAP *bmp;
 
-   bmp = al_load_bitmap(filename);
+   bmp = al_load_bitmap_flags(filename, flags);
    if (!bmp) {
-      error("failed to load %s", filename);
+      fprintf(stderr, "test_driver: failed to load %s\n", filename);
+      bmp = create_fallback_bitmap();
    }
    return bmp;
 }
 
 static void load_bitmaps(ALLEGRO_CONFIG const *cfg, const char *section,
-   BmpType bmp_type)
+   BmpType bmp_type, int flags)
 {
    int i = 0;
    ALLEGRO_CONFIG_ENTRY *iter;
@@ -157,14 +184,14 @@ static void load_bitmaps(ALLEGRO_CONFIG const *cfg, const char *section,
       value = al_get_config_value(cfg, section, key);
 
       bitmaps[i].name = al_ustr_new(key);
-      bitmaps[i].bitmap[bmp_type] = load_relative_bitmap(value);
+      bitmaps[i].bitmap[bmp_type] = load_relative_bitmap(value, flags);
 
       key = al_get_next_config_entry(&iter);
       i++;
    }
 
    if (i == MAX_BITMAPS)
-      error("bitmap limit reached");
+      fatal_error("bitmap limit reached");
 
    num_global_bitmaps = i;
 }
@@ -180,7 +207,7 @@ static ALLEGRO_BITMAP **reserve_local_bitmap(const char *name, BmpType bmp_type)
       }
    }
 
-   error("bitmap limit reached");
+   fatal_error("bitmap limit reached");
    return NULL;
 }
 
@@ -214,6 +241,8 @@ static void set_target_reset(ALLEGRO_BITMAP *target)
    al_set_blender(ALLEGRO_ADD, ALLEGRO_ONE, ALLEGRO_ZERO);
    al_identity_transform(&ident);
    al_use_transform(&ident);
+   al_orthographic_transform(&ident, 0, 0, -1, al_get_bitmap_width(target), al_get_bitmap_height(target), 1);
+   al_use_projection_transform(&ident);
 }
 
 static char const *resolve_var(ALLEGRO_CONFIG const *cfg, char const *section,
@@ -257,10 +286,19 @@ static ALLEGRO_BITMAP *get_bitmap(char const *value, BmpType bmp_type,
    if (streq(value, "0") || streq(value, "NULL"))
       return NULL;
 
-   error("undefined bitmap: %s", value);
+   fatal_error("undefined bitmap: %s", value);
    return NULL;
 }
 
+static int get_load_bitmap_flag(char const *value)
+{
+   if (streq(value, "ALLEGRO_NO_PREMULTIPLIED_ALPHA"))
+      return ALLEGRO_NO_PREMULTIPLIED_ALPHA;
+   if (streq(value, "ALLEGRO_KEEP_INDEX"))
+      return ALLEGRO_KEEP_INDEX;
+   return atoi(value);
+}
+
 static int get_draw_bitmap_flag(char const *value)
 {
    if (streq(value, "ALLEGRO_FLIP_HORIZONTAL"))
@@ -292,6 +330,8 @@ static int get_blend_factor(char const *value)
       : streq(value, "ALLEGRO_DEST_COLOR") ? ALLEGRO_DEST_COLOR
       : streq(value, "ALLEGRO_INVERSE_SRC_COLOR") ? ALLEGRO_INVERSE_SRC_COLOR
       : streq(value, "ALLEGRO_INVERSE_DEST_COLOR") ? ALLEGRO_INVERSE_DEST_COLOR
+      : streq(value, "ALLEGRO_CONST_COLOR") ? ALLEGRO_CONST_COLOR
+      : streq(value, "ALLEGRO_INVERSE_CONST_COLOR") ? ALLEGRO_INVERSE_CONST_COLOR
       : atoi(value);
 }
 
@@ -310,7 +350,7 @@ static ALLEGRO_TRANSFORM *get_transform(const char *name)
          return &transforms[i].transform;
    }
 
-   error("transforms limit reached");
+   fatal_error("transforms limit reached");
    return NULL;
 }
 
@@ -343,9 +383,12 @@ static int get_pixel_format(char const *v)
       : streq(v, "ALLEGRO_PIXEL_FORMAT_ABGR_F32") ? ALLEGRO_PIXEL_FORMAT_ABGR_F32
       : streq(v, "ALLEGRO_PIXEL_FORMAT_ABGR_8888_LE") ? ALLEGRO_PIXEL_FORMAT_ABGR_8888_LE
       : streq(v, "ALLEGRO_PIXEL_FORMAT_RGBA_4444") ? ALLEGRO_PIXEL_FORMAT_RGBA_4444
+      : streq(v, "ALLEGRO_PIXEL_FORMAT_COMPRESSED_RGBA_DXT1") ? ALLEGRO_PIXEL_FORMAT_COMPRESSED_RGBA_DXT1
+      : streq(v, "ALLEGRO_PIXEL_FORMAT_COMPRESSED_RGBA_DXT3") ? ALLEGRO_PIXEL_FORMAT_COMPRESSED_RGBA_DXT3
+      : streq(v, "ALLEGRO_PIXEL_FORMAT_COMPRESSED_RGBA_DXT5") ? ALLEGRO_PIXEL_FORMAT_COMPRESSED_RGBA_DXT5
       : -1;
    if (format == -1)
-      error("invalid format: %s", v);
+      fatal_error("invalid format: %s", v);
    return format;
 }
 
@@ -357,6 +400,13 @@ static int get_lock_bitmap_flags(char const *v)
       : atoi(v);
 }
 
+static int get_bitmap_flags(char const *v)
+{
+   return streq(v, "ALLEGRO_MEMORY_BITMAP") ? ALLEGRO_MEMORY_BITMAP
+      : streq(v, "ALLEGRO_VIDEO_BITMAP") ? ALLEGRO_VIDEO_BITMAP
+      : atoi(v);
+}
+
 static void fill_lock_region(LockRegion *lr, float alphafactor, bool blended)
 {
    int x, y;
@@ -425,7 +475,7 @@ static void load_fonts(ALLEGRO_CONFIG const *cfg, const char *section)
 
       if (load_stmt) {
          if (!font) {
-            error("failed to load font: %s", key);
+            fatal_error("failed to load font: %s", key);
          }
          fonts[i].name = al_ustr_new(key);
          fonts[i].font = font;
@@ -436,7 +486,7 @@ static void load_fonts(ALLEGRO_CONFIG const *cfg, const char *section)
    }
 
    if (i == MAX_FONTS)
-      error("font limit reached");
+      fatal_error("font limit reached");
 
 #undef MAXBUF
 }
@@ -445,12 +495,15 @@ static ALLEGRO_FONT *get_font(char const *name)
 {
    int i;
 
+   if (streq(name, "NULL"))
+      return NULL;
+
    for (i = 0; i < MAX_FONTS; i++) {
       if (fonts[i].name && streq(al_cstr(fonts[i].name), name))
          return fonts[i].font;
    }
 
-   error("undefined font: %s", name);
+   fatal_error("undefined font: %s", name);
    return NULL;
 }
 
@@ -459,6 +512,13 @@ static int get_font_align(char const *value)
    return streq(value, "ALLEGRO_ALIGN_LEFT") ? ALLEGRO_ALIGN_LEFT
       : streq(value, "ALLEGRO_ALIGN_CENTRE") ? ALLEGRO_ALIGN_CENTRE
       : streq(value, "ALLEGRO_ALIGN_RIGHT") ? ALLEGRO_ALIGN_RIGHT
+      : streq(value, "ALLEGRO_ALIGN_INTEGER") ? ALLEGRO_ALIGN_INTEGER
+      : streq(value, "ALLEGRO_ALIGN_LEFT|ALLEGRO_ALIGN_INTEGER")
+         ? ALLEGRO_ALIGN_LEFT | ALLEGRO_ALIGN_INTEGER
+      : streq(value, "ALLEGRO_ALIGN_RIGHT|ALLEGRO_ALIGN_INTEGER")
+         ? ALLEGRO_ALIGN_RIGHT | ALLEGRO_ALIGN_INTEGER
+      : streq(value, "ALLEGRO_ALIGN_CENTRE|ALLEGRO_ALIGN_INTEGER")
+         ? ALLEGRO_ALIGN_CENTRE | ALLEGRO_ALIGN_INTEGER
       : atoi(value);
 }
 
@@ -470,6 +530,14 @@ static void set_config_int(ALLEGRO_CONFIG *cfg, char const *section,
    al_set_config_value(cfg, section, var, buf);
 }
 
+static void set_config_float(ALLEGRO_CONFIG *cfg, char const *section,
+   char const *var, float value)
+{
+   char buf[40];
+   sprintf(buf, "%f", value);
+   al_set_config_value(cfg, section, var, buf);
+}
+
 static void fill_vertices(ALLEGRO_CONFIG const *cfg, char const *name)
 {
 #define MAXBUF    80
@@ -502,6 +570,59 @@ static void fill_vertices(ALLEGRO_CONFIG const *cfg, char const *name)
 #undef MAXBUF
 }
 
+static void fill_simple_vertices(ALLEGRO_CONFIG const *cfg, char const *name)
+{
+#define MAXBUF    80
+
+   char const *value;
+   char buf[MAXBUF];
+   float x, y;
+   int i;
+
+   memset(simple_vertices, 0, sizeof(simple_vertices));
+
+   for (i = 0; i < MAX_VERTICES; i++) {
+      sprintf(buf, "v%d", i);
+      value = al_get_config_value(cfg, name, buf);
+      if (!value)
+         break;
+
+      if (sscanf(value, " %f , %f", &x, &y) == 2) {
+         simple_vertices[2*i + 0] = x;
+         simple_vertices[2*i + 1] = y;
+      }
+   }
+
+   num_simple_vertices = i;
+
+#undef MAXBUF
+}
+
+static void fill_vertex_counts(ALLEGRO_CONFIG const *cfg, char const *name)
+{
+#define MAXBUF    80
+
+   char const *value;
+   char buf[MAXBUF];
+   int count;
+   int i;
+
+   memset(vertex_counts, 0, sizeof(vertex_counts));
+
+   for (i = 0; i < MAX_POLYGONS; i++) {
+      sprintf(buf, "p%d", i);
+      value = al_get_config_value(cfg, name, buf);
+      if (!value)
+         break;
+
+      if (sscanf(value, " %d ", &count) == 1) {
+         vertex_counts[i] = count;
+      }
+   }
+
+#undef MAXBUF
+}
+
 static int get_prim_type(char const *value)
 {
    return streq(value, "ALLEGRO_PRIM_POINT_LIST") ? ALLEGRO_PRIM_POINT_LIST
@@ -514,6 +635,25 @@ static int get_prim_type(char const *value)
       : atoi(value);
 }
 
+static int get_line_join(char const *value)
+{
+   return streq(value, "ALLEGRO_LINE_JOIN_NONE") ? ALLEGRO_LINE_JOIN_NONE
+      : streq(value, "ALLEGRO_LINE_JOIN_BEVEL") ? ALLEGRO_LINE_JOIN_BEVEL
+      : streq(value, "ALLEGRO_LINE_JOIN_ROUND") ? ALLEGRO_LINE_JOIN_ROUND
+      : streq(value, "ALLEGRO_LINE_JOIN_MITER") ? ALLEGRO_LINE_JOIN_MITER
+      : atoi(value);
+}
+
+static int get_line_cap(char const *value)
+{
+   return streq(value, "ALLEGRO_LINE_CAP_NONE") ? ALLEGRO_LINE_CAP_NONE
+      : streq(value, "ALLEGRO_LINE_CAP_SQUARE") ? ALLEGRO_LINE_CAP_SQUARE
+      : streq(value, "ALLEGRO_LINE_CAP_ROUND") ? ALLEGRO_LINE_CAP_ROUND
+      : streq(value, "ALLEGRO_LINE_CAP_TRIANGLE") ? ALLEGRO_LINE_CAP_TRIANGLE
+      : streq(value, "ALLEGRO_LINE_CAP_CLOSED") ? ALLEGRO_LINE_CAP_CLOSED
+      : atoi(value);
+}
+
 /* FNV-1a algorithm, parameters from:
  * http://www.isthe.com/chongo/tech/comp/fnv/index.html
  */
@@ -588,7 +728,7 @@ static int base64_decode(char c)
    if (c >= 'a' && c <= 'z') return 36 + (c - 'a');
    if (c == '+') return 62;
    if (c == '/') return 63;
-   error("invalid base64 character: %c", c);
+   fatal_error("invalid base64 character: %c", c);
    return -1;
 }
 
@@ -734,13 +874,28 @@ static void check_similarity(ALLEGRO_CONFIG const *cfg,
    
    if (bmp_type == HW) {
       char const *exp = al_get_config_value(cfg, testname, "hash_hw");
+      char const *sigexp = al_get_config_value(cfg, testname, "sig_hw");
       char hash[16];
+      char sig[SIG_LENZ];
       sprintf(hash, "%08x", hash_bitmap(bmp1));
       if (exp && streq(hash, exp)) {
          printf("OK   %s [%s]\n", testname, bt);
          passed_tests++;
          return;
       }
+
+      if (sigexp && strlen(sigexp) != SIG_LEN) {
+         printf("WARNING: ignoring bad signature: %s\n", sigexp);
+         sigexp = NULL;
+      }
+
+      compute_signature(bmp1, sig);
+
+      if (sigexp && similar_signatures(sig, sigexp)) {
+         printf("OK   %s [%s] - by signature\n", testname, bt);
+         passed_tests++;
+         return;
+      }
    }
 
    /* The default cutoff is "empirically determined" only. */
@@ -845,11 +1000,27 @@ static void do_test(ALLEGRO_CONFIG *cfg, char const *testname,
          continue;
       }
 
+      if (SCAN("al_set_new_bitmap_flags", 1)) {
+         al_set_new_bitmap_flags(get_bitmap_flags(V(0)));
+         continue;
+      }
+
+      if (SCAN("al_set_new_bitmap_format", 1)) {
+         al_set_new_bitmap_format(get_pixel_format(V(0)));
+         continue;
+      }
+
       if (SCAN("al_clear_to_color", 1)) {
          al_clear_to_color(C(0));
          continue;
       }
 
+      if (SCANLVAL("al_clone_bitmap", 1)) {
+         ALLEGRO_BITMAP **bmp = reserve_local_bitmap(lval, bmp_type);
+         (*bmp) = al_clone_bitmap(B(0));
+         continue;
+      }
+
       if (SCAN("al_draw_bitmap", 4)) {
          al_draw_bitmap(B(0), F(1), F(2),
             get_draw_bitmap_flag(V(3)));
@@ -954,15 +1125,27 @@ static void do_test(ALLEGRO_CONFIG *cfg, char const *testname,
 
       if (SCANLVAL("al_load_bitmap", 1)) {
          ALLEGRO_BITMAP **bmp = reserve_local_bitmap(lval, bmp_type);
-         (*bmp) = load_relative_bitmap(V(0));
+         (*bmp) = load_relative_bitmap(V(0), 0);
+         continue;
+      }
+      if (SCANLVAL("al_load_bitmap_flags", 2)) {
+         ALLEGRO_BITMAP **bmp = reserve_local_bitmap(lval, bmp_type);
+         (*bmp) = load_relative_bitmap(V(0), get_load_bitmap_flag(V(1)));
          continue;
       }
       if (SCAN("al_save_bitmap", 2)) {
          if (!al_save_bitmap(V(0), B(1))) {
-            error("failed to save %s", V(0));
+            fatal_error("failed to save %s", V(0));
          }
          continue;
       }
+      if (SCANLVAL("al_identify_bitmap", 1)) {
+         char const *ext = al_identify_bitmap(V(0));
+         if (!ext)
+            ext = "NULL";
+         al_set_config_value(cfg, testname, lval, ext);
+         continue;
+      }
 
       if (SCAN("al_hold_bitmap_drawing", 1)) {
          al_hold_bitmap_drawing(get_bool(V(0)));
@@ -999,6 +1182,13 @@ static void do_test(ALLEGRO_CONFIG *cfg, char const *testname,
          continue;
       }
 
+      /* Conversion */
+      if (SCAN("al_convert_bitmap", 1)) {
+         ALLEGRO_BITMAP *bmp = B(0);
+         al_convert_bitmap(bmp);
+         continue;
+      }
+
       /* Locking */
       if (SCAN("al_lock_bitmap", 3)) {
          ALLEGRO_BITMAP *bmp = B(0);
@@ -1074,6 +1264,10 @@ static void do_test(ALLEGRO_CONFIG *cfg, char const *testname,
          set_config_int(cfg, testname, V(5), bbh);
          continue;
       }
+      if (SCAN("al_set_fallback_font", 2)) {
+         al_set_fallback_font(get_font(V(0)), get_font(V(1)));
+         continue;
+      }
 
       /* Primitives */
       if (SCAN("al_draw_line", 6)) {
@@ -1154,7 +1348,138 @@ static void do_test(ALLEGRO_CONFIG *cfg, char const *testname,
          continue;
       }
 
-      error("statement didn't scan: %s", stmt);
+      /* Keep 5.0 and 5.1 functions separate for easier merging. */
+
+      /* Primitives (5.1) */
+      if (SCAN("al_draw_polyline", 6)) {
+         fill_simple_vertices(cfg, V(0));
+         al_draw_polyline(simple_vertices, 2 * sizeof(float),
+            num_simple_vertices, get_line_join(V(1)), get_line_cap(V(2)),
+            C(3), F(4), F(5));
+         continue;
+      }
+      if (SCAN("al_draw_polygon", 5)) {
+         fill_simple_vertices(cfg, V(0));
+         al_draw_polygon(simple_vertices, num_simple_vertices,
+            get_line_join(V(1)), C(2), F(3), F(4));
+         continue;
+      }
+      if (SCAN("al_draw_filled_polygon", 2)) {
+         fill_simple_vertices(cfg, V(0));
+         al_draw_filled_polygon(simple_vertices, num_simple_vertices, C(1));
+         continue;
+      }
+      if (SCAN("al_draw_filled_polygon_with_holes", 3)) {
+         fill_simple_vertices(cfg, V(0));
+         fill_vertex_counts(cfg, V(1));
+         al_draw_filled_polygon_with_holes(simple_vertices, vertex_counts, C(2));
+         continue;
+      }
+
+      /* Transformations (5.1) */
+      if (SCAN("al_horizontal_shear_transform", 2)) {
+         al_horizontal_shear_transform(get_transform(V(0)), F(1));
+         continue;
+      }
+      if (SCAN("al_vertical_shear_transform", 2)) {
+         al_vertical_shear_transform(get_transform(V(0)), F(1));
+         continue;
+      }
+      if (SCAN("al_orthographic_transform", 7)) {
+         al_orthographic_transform(get_transform(V(0)), F(1), F(2), F(3), F(4), F(5), F(6));
+         continue;
+      }
+      if (SCAN("al_use_projection_transform", 1)) {
+         al_use_projection_transform(get_transform(V(0)));
+         continue;
+      }
+      if (SCAN("al_set_blend_color", 1)) {
+         al_set_blend_color(C(0));
+         continue;
+      }
+
+      /* Simple arithmetic, generally useful. (5.1) */
+      if (SCANLVAL("isum", 2)) {
+         int result  = I(0) + I(1);
+         set_config_int(cfg, testname, lval, result);
+         continue;
+      }
+      if (SCANLVAL("idif", 2)) {
+         int result  = I(0) - I(1);
+         set_config_int(cfg, testname, lval, result);
+         continue;
+      }
+      if (SCANLVAL("imul", 2)) {
+         int result  = I(0) * I(1);
+         set_config_int(cfg, testname, lval, result);
+         continue;
+      }
+      if (SCANLVAL("idiv", 2)) {
+         int result  = I(0) / I(1);
+         set_config_int(cfg, testname, lval, result);
+         continue;
+      }
+      if (SCANLVAL("fsum", 2)) {
+         float result  = F(0) + F(1);
+         set_config_float(cfg, testname, lval, result);
+         continue;
+      }
+      if (SCANLVAL("fdif", 2)) {
+         float result  = F(0) - F(1);
+         set_config_float(cfg, testname, lval, result);
+         continue;
+      }
+      if (SCANLVAL("fmul", 2)) {
+         float result  = F(0) * F(1);
+         set_config_float(cfg, testname, lval, result);
+         continue;
+      }
+      if (SCANLVAL("fdiv", 2)) {
+         float result  = F(0) / F(1);
+         set_config_float(cfg, testname, lval, result);
+         continue;
+      }
+
+      /* Dynamical variable initialisation, needed to properly initialize
+       * variables (5.1)*/
+      if (SCANLVAL("int", 1)) {
+         float result  = F(0);
+         set_config_float(cfg, testname, lval, result);
+         continue;
+      }
+      if (SCANLVAL("float", 1)) {
+         float result  = F(0);
+         set_config_float(cfg, testname, lval, result);
+         continue;
+      }
+       
+      /* Fonts: per glyph text handling (5.1)  */
+      if (SCAN("al_draw_glyph", 5)) {
+         al_draw_glyph(get_font(V(0)), C(1), F(2), F(3), I(4));
+         continue;
+      }
+      if (SCANLVAL("al_get_glyph_advance", 3)) {
+         int kerning = al_get_glyph_advance(get_font(V(0)), I(1), I(2));
+         set_config_int(cfg, testname, lval, kerning);
+         continue;
+      }
+      if (SCANLVAL("al_get_glyph_width", 2)) {
+         int w = al_get_glyph_width(get_font(V(0)), I(1));
+         set_config_int(cfg, testname, lval, w);
+         continue;
+      }
+      if (SCANLVAL("al_get_glyph_dimensions", 6)) {
+         int bbx, bby, bbw, bbh;
+         bool ok = al_get_glyph_dimensions(get_font(V(0)), I(1), &bbx, &bby, &bbw, &bbh);
+         set_config_int(cfg, testname, V(2), bbx);
+         set_config_int(cfg, testname, V(3), bby);
+         set_config_int(cfg, testname, V(4), bbw);
+         set_config_int(cfg, testname, V(5), bbh);
+         set_config_int(cfg, testname, lval, ok);
+         continue;
+      }
+
+      fatal_error("statement didn't scan: %s", stmt);
    }
 
    al_set_new_bitmap_format(ALLEGRO_PIXEL_FORMAT_ANY_WITH_ALPHA);
@@ -1214,13 +1539,13 @@ static void sw_hw_test(ALLEGRO_CONFIG *cfg, char const *testname)
    int old_failed_tests = failed_tests;
    bool reliable;
 
-   al_set_new_bitmap_flags(ALLEGRO_MEMORY_BITMAP | ALLEGRO_NO_PREMULTIPLIED_ALPHA);
+   al_set_new_bitmap_flags(ALLEGRO_MEMORY_BITMAP);
    do_test(cfg, testname, membuf, SW, true);
 
    reliable = (failed_tests == old_failed_tests);
 
    if (display) {
-      al_set_new_bitmap_flags(ALLEGRO_VIDEO_BITMAP | ALLEGRO_NO_PREMULTIPLIED_ALPHA);
+      al_set_new_bitmap_flags(ALLEGRO_VIDEO_BITMAP);
       do_test(cfg, testname, al_get_backbuffer(display), HW, reliable);
    }
 }
@@ -1243,7 +1568,7 @@ static void merge_config_sections(
    value = al_get_config_value(src_cfg, src_section, "extend");
    if (value) {
       if (streq(value, src_section)) {
-         error("section cannot extend itself: %s "
+         fatal_error("section cannot extend itself: %s "
             "(did you forget to rename a section?)", src_section);
       }
       merge_config_sections(targ_cfg, targ_section, src_cfg, value);
@@ -1251,7 +1576,7 @@ static void merge_config_sections(
 
    key = al_get_first_config_entry(src_cfg, src_section, &iter);
    if (!key) {
-      error("missing section: %s", src_section);
+      fatal_error("missing section: %s", src_section);
    }
    for (; key != NULL; key = al_get_next_config_entry(&iter)) {
       value = al_get_config_value(src_cfg, src_section, key);
@@ -1265,7 +1590,7 @@ static void run_test(ALLEGRO_CONFIG const *cfg, char const *section)
    ALLEGRO_CONFIG *cfg2;
 
    if (!section_exists(cfg, section)) {
-      error("section not found: %s", section);
+      fatal_error("section not found: %s", section);
    }
 
    cfg2 = al_create_config();
@@ -1336,10 +1661,10 @@ static void process_ini_files(void)
 
    while (argc > 0) {
       if (!has_suffix(argv[0], ".ini"))
-         error("expected .ini argument: %s\n", argv[0]);
+         fatal_error("expected .ini argument: %s\n", argv[0]);
       cfg = al_load_config_file(argv[0]);
       if (!cfg)
-         error("failed to load config file %s", argv[0]);
+         fatal_error("failed to load config file %s", argv[0]);
 
       if (verbose)
          printf("Running %s\n", argv[0]);
@@ -1347,12 +1672,12 @@ static void process_ini_files(void)
       argc--;
       argv++;
 
-      al_set_new_bitmap_flags(ALLEGRO_MEMORY_BITMAP | ALLEGRO_NO_PREMULTIPLIED_ALPHA);
-      load_bitmaps(cfg, "bitmaps", SW);
+      al_set_new_bitmap_flags(ALLEGRO_MEMORY_BITMAP);
+      load_bitmaps(cfg, "bitmaps", SW, ALLEGRO_NO_PREMULTIPLIED_ALPHA);
 
       if (display) {
-         al_set_new_bitmap_flags(ALLEGRO_VIDEO_BITMAP | ALLEGRO_NO_PREMULTIPLIED_ALPHA);
-         load_bitmaps(cfg, "bitmaps", HW);
+         al_set_new_bitmap_flags(ALLEGRO_VIDEO_BITMAP);
+         load_bitmaps(cfg, "bitmaps", HW, ALLEGRO_NO_PREMULTIPLIED_ALPHA);
       }
 
       load_fonts(cfg, "fonts");
@@ -1373,19 +1698,41 @@ static void process_ini_files(void)
    }
 }
 
+const char* help_str =
+" [OPTION] CONFIG_FILE [TEST_NAME]... [CONFIG_FILE [TEST_NAME]...]...\n"
+"\n"
+"Run Allegro graphical output tests within one or more CONFIG_FILEs (each\n"
+"having an .ini extension). By default this program runs all the tests in a\n"
+"file, but individual TEST_NAMEs can be specified after each CONFIG_FILE.\n"
+"\n"
+"Options:\n"
+" -d, --delay        duration (in sec) to wait between tests\n"
+" --force-d3d        force using D3D (Windows only)\n"
+" --force-opengl-1.2 force using OpenGL 1.2\n"
+" --force-opengl-2.0 force using OpenGL 2.0\n"
+" --force-opengl     force using OpenGL\n"
+" -h, --help         display this message\n"
+" -n, --no-display   do not create a display (hardware drawing is disabled)\n"
+" -s, --save         save the output of each test in the current directory\n"
+" --use-shaders      use the programmable pipeline for drawing\n"
+" -v, --verbose      show additional information after each test\n"
+" -q, --quiet        do not draw test output to the display\n";
+
 int main(int _argc, char *_argv[])
 {
+   int display_flags = 0;
+
    argc = _argc;
    argv = _argv;
 
    if (argc == 1) {
-      error("requires config file argument");
+      fatal_error("requires config file argument.\nSee --help for usage");
    }
    argc--;
    argv++;
 
    if (!al_init()) {
-      error("failed to initialise Allegro");
+      fatal_error("failed to initialise Allegro");
    }
    al_init_image_addon();
    al_init_font_addon();
@@ -1413,19 +1760,26 @@ int main(int _argc, char *_argv[])
       else if (streq(opt, "--force-opengl-1.2")) {
          ALLEGRO_CONFIG *cfg = al_get_system_config();
          al_set_config_value(cfg, "opengl", "force_opengl_version", "1.2");
-         al_set_new_display_flags(ALLEGRO_OPENGL);
+         display_flags |= ALLEGRO_OPENGL;
       }
       else if (streq(opt, "--force-opengl-2.0")) {
          ALLEGRO_CONFIG *cfg = al_get_system_config();
          al_set_config_value(cfg, "opengl", "force_opengl_version", "2.0");
-         al_set_new_display_flags(ALLEGRO_OPENGL);
+         display_flags |= ALLEGRO_OPENGL;
       }
       else if (streq(opt, "--force-opengl")) {
-         al_set_new_display_flags(ALLEGRO_OPENGL);
+         display_flags |= ALLEGRO_OPENGL;
       }
       else if (streq(opt, "--force-d3d")) {
          /* Don't try this at home. */
-         al_set_new_display_flags(ALLEGRO_DIRECT3D_INTERNAL);
+         display_flags |= ALLEGRO_DIRECT3D_INTERNAL;
+      }
+      else if (streq(opt, "--use-shaders")) {
+         display_flags |= ALLEGRO_PROGRAMMABLE_PIPELINE;
+      }
+      else if (streq(opt, "-h") || streq(opt, "--help")) {
+         printf("Usage:\n%s%s", _argv[0], help_str);
+         return 0;
       }
       else {
          break;
@@ -1433,9 +1787,10 @@ int main(int _argc, char *_argv[])
    }
 
    if (want_display) {
+      al_set_new_display_flags(display_flags);
       display = al_create_display(640, 480);
       if (!display) {
-         error("failed to create display");
+         fatal_error("failed to create display");
       }
    }
 
